from typing import Dict, List, Union

from docarray import Document, DocumentArray
from docarray.score import NamedScore
from numpy import dot
from numpy.linalg import norm

from now.utils.docarray.helpers import get_chunk_by_field_name


def convert_es_to_da(
    result: Union[Dict, List[Dict]], get_score_breakdown: bool
) -> DocumentArray:
    """
    Transform Elasticsearch documents into DocumentArray. Assumes that all Elasticsearch
    documents have a 'text' field. It returns embeddings as part of the tags for each field that is encoded.

    :param result: results from an Elasticsearch query.
    :param get_score_breakdown: whether to return the embeddings as tags for each document.
    :return: a DocumentArray containing all results.
    """
    if isinstance(result, Dict):
        result = [result]
    da = DocumentArray()
    for es_doc in result:
        doc = Document.from_base64(es_doc['_source']['serialized_doc'])
        for k, v in es_doc['_source'].items():
            if (
                k.startswith('embedding') or k.endswith('embedding')
            ) and get_score_breakdown:
                if 'embeddings' not in doc.tags:
                    doc.tags['embeddings'] = {}
                doc.tags['embeddings'][k] = v
        da.append(doc)
    return da


def convert_doc_map_to_es(
    docs_map: Dict[str, DocumentArray],
    index_name: str,
    encoder_to_fields: dict,
) -> List[Dict]:
    """
    Transform a dictionary (mapping encoder to DocumentArray) into a list of Elasticsearch documents.
    The `docs_map` dictionary is expected to have the following structure:
    {
        'encoder1': DocumentArray([...]),
        'encoder2': DocumentArray([...]), # same number of documents as encoder1
        ...
    }

    :param docs_map: dictionary mapping encoder to DocumentArray.
    :param index_name: name of the index to be used in Elasticsearch.
    :param encoder_to_fields: dictionary mapping encoder to fields.
    :return: a list of Elasticsearch documents as dictionaries ready to be indexed.
    """
    es_docs = {}
    for executor_name, documents in docs_map.items():
        for doc in documents:
            if doc.id not in es_docs:
                es_docs[doc.id] = get_base_es_doc(doc, index_name)
                _doc = DocumentArray(Document(doc, copy=True))
                # remove embeddings from serialized doc
                _doc[..., 'embedding'] = None
                es_docs[doc.id]['serialized_doc'] = _doc[0].to_base64()
            es_doc = es_docs[doc.id]
            for encoded_field in encoder_to_fields[executor_name]:
                field_doc = get_chunk_by_field_name(doc, encoded_field)
                es_doc[
                    f'{encoded_field}-{executor_name}.embedding'
                ] = field_doc.embedding
                if hasattr(field_doc, 'text') and field_doc.text:
                    es_doc[f'{encoded_field}'] = field_doc.text
                if hasattr(field_doc, 'uri') and field_doc.uri:
                    es_doc['uri'] = field_doc.uri
    return list(es_docs.values())


def get_base_es_doc(doc: Document, index_name: str) -> Dict:
    es_doc = {k: v for k, v in doc.to_dict().items() if v}
    es_doc.pop('chunks', None)
    es_doc.pop('_metadata', None)
    es_doc['_op_type'] = 'index'
    es_doc['_index'] = index_name
    es_doc['_id'] = doc.id
    # TODO remove side effect - should not be part of this function
    doc.tags['embeddings'] = {}
    return es_doc


def convert_es_results_to_matches(
    query_doc: Document,
    es_results: List[Dict],
    get_score_breakdown: bool,
    metric: str,
    score_calculation,
) -> DocumentArray:
    """
    Transform a list of results from Elasticsearch into a matches in the form of a `DocumentArray`.

    :param query_doc: the query document.
    :param es_results: List of dictionaries containing results from Elasticsearch querying.
    :param get_score_breakdown: whether to calculate the score breakdown for matches.
    :param metric: the metric used to calculate the score.
    :param score_calculation: the score calculation for each match.

    :return: `DocumentArray` that holds all matches in the form of `Document`s.
    """
    matches = DocumentArray()
    for result in es_results:
        d = convert_es_to_da(result, get_score_breakdown)[0]
        d.scores[metric] = NamedScore(value=result['_score'])
        if get_score_breakdown:
            d = calculate_score_breakdown(query_doc, d, score_calculation, metric)
        d.embedding = None
        matches.append(d)
    return matches


def calculate_score_breakdown(
    query_doc: Document, retrieved_doc: Document, score_calculation, metric
) -> Document:
    """
    Calculate the score breakdown for a given retrieved document. Each score calculation in the indexer's
    `score_calculation` should have a corresponding value, returned inside a list of scores in the documents
    tags under `score_breakdown`.

    :param query_doc: The query document. Contains embeddings for the score calculation at tag level.
    :param retrieved_doc: The Elasticsearch results, containing embeddings inside the `_source` field.
    :param score_calculation: The score calculation used for the score breakdown.
    :param metric: The metric to be used for the score breakdown.

    :return: List of integers representing the score breakdown.
    """
    retrieved_doc.scores['total'] = retrieved_doc.scores.pop(
        metric
    )  # save the final script score as total
    add_bm25 = False
    for (
        query_field,
        document_field,
        encoder,
        linear_weight,
    ) in score_calculation:
        if encoder == 'bm25':
            add_bm25 = True
            continue
        q_emb = query_doc.tags['embeddings'][f'{query_field}-{encoder}']
        d_emb = retrieved_doc.tags['embeddings'][
            f'{document_field}-{encoder}.embedding'
        ]
        if metric == 'cosine':
            score = calculate_cosine(d_emb, q_emb) * linear_weight
        elif metric == 'l2_norm':
            score = calculate_l2_norm(d_emb, q_emb) * linear_weight
        else:
            raise ValueError(f'Invalid metric {metric}')
        retrieved_doc.scores[
            '-'.join(
                [
                    query_field,
                    document_field,
                    encoder,
                    str(linear_weight),
                ]
            )
        ] = NamedScore(value=round(score, 6))

    if add_bm25:
        # calculate bm25 score
        vector_total = sum(
            [v.value for k, v in retrieved_doc.scores.items() if k != 'total']
        )
        bm25_normalized = retrieved_doc.scores['total'].value - vector_total - 1
        bm25_raw = bm25_normalized * 10
        retrieved_doc.scores['bm25_normalized'] = NamedScore(
            value=round(bm25_normalized, 6)
        )
        retrieved_doc.scores['bm25_raw'] = NamedScore(value=round(bm25_raw, 6))

    # remove embeddings from document
    retrieved_doc.tags.pop('embeddings', None)
    return retrieved_doc


def calculate_l2_norm(d_emb, q_emb):
    return norm(q_emb - d_emb)


def calculate_cosine(d_emb, q_emb):
    return dot(q_emb, d_emb) / (norm(q_emb) * norm(d_emb))



import datetime
import json
import logging
import os
import threading
from time import sleep

from hubble.payment.client import PaymentClient

from now.constants import (
    NOWGATEWAY_BASE_FEE_QUANTITY,
    NOWGATEWAY_BASE_FEE_SLEEP_INTERVAL,
    NOWGATEWAY_SEARCH_FEE_PRO_QUANTITY,
    NOWGATEWAY_SEARCH_FEE_QUANTITY,
)

logger = logging.getLogger(__file__)
logger.setLevel(os.environ.get('JINA_LOG_LEVEL', 'INFO'))
payment_client = None
authorized_jwt = None
old_user_token = None


def current_time():
    return datetime.datetime.utcnow().isoformat() + 'Z'


def start_base_fee_thread(user_token, inf_token, storage_dir):
    logger.info('Starting base fee thread')
    global old_user_token, authorized_jwt
    old_user_token = user_token  # incase no token is passed with search request
    if inf_token:  # if inf_token is passed, use it mostly in case of gateway restart
        authorized_jwt = inf_token
    init_payment_client(user_token)
    save_cred(storage_dir)
    thread = threading.Thread(target=base_fee_thread, args=(user_token,))
    thread.start()


def save_cred(storage_dir):
    if storage_dir:
        with open(f'{storage_dir}/cred.json', 'w') as f:
            json.dump({'authorized_jwt': authorized_jwt}, f)
    else:
        logger.info('No storage dir found. Not saving cred.json')


def base_fee_thread(user_token):
    while True:
        sleep(NOWGATEWAY_BASE_FEE_SLEEP_INTERVAL)
        report(
            user_token=user_token,
            quantity_basic=NOWGATEWAY_BASE_FEE_QUANTITY,
            quantity_pro=NOWGATEWAY_BASE_FEE_QUANTITY,
        )


def report_search_usage(user_token):
    logger.info('** Entered report_search_usage() **')
    init_payment_client(user_token)
    report(
        user_token=user_token,
        quantity_basic=NOWGATEWAY_SEARCH_FEE_QUANTITY,
        quantity_pro=NOWGATEWAY_SEARCH_FEE_PRO_QUANTITY,
    )


def init_payment_client(user_token):
    global payment_client
    global authorized_jwt
    try:
        if payment_client is None:
            m2m_token = os.environ.get('M2M')
            if not m2m_token:
                raise ValueError('M2M not set in the environment')
            logger.info(f'M2M_TOKEN: {m2m_token[:10]}...{m2m_token[-10:]}')
            payment_client = PaymentClient(m2m_token=m2m_token)
        if authorized_jwt is None:
            logger.info('No authorized JWT found. Getting one...')
            # Try with the new token provided. If not then use the old one
            jwt = payment_client.get_authorized_jwt(
                user_token=user_token or old_user_token
            )['data']
            authorized_jwt = jwt  # store it as a global variable to avoid re-auth
    except Exception as e:
        import traceback

        traceback.print_exc()
        # Do not continue with request if initialization fails
        raise e


def report(user_token, quantity_basic, quantity_pro):
    logger.info('Time of report: {}'.format(current_time()))
    app_id = 'search'
    product_id = 'free-plan'
    try:
        logger.info(f'Charging user with token {user_token or old_user_token}')
        logger.info(f'Authorized JWT: {authorized_jwt[:10]}...{authorized_jwt[-10:]}')
        summary = get_summary()
        logger.info(f'Credits before: {summary["credits"]}')
        if summary['internal_product_id'] == 'free-plan':
            quantity = quantity_basic
        else:
            quantity = quantity_pro
        if can_charge(summary):
            payment_client.report_usage(authorized_jwt, app_id, product_id, quantity)  # type: ignore
            logger.info(f'**** `{round(quantity, 3)}` credits charged ****')
            summary = get_summary()
            logger.info(f'Credits after: {summary["credits"]}')
        else:
            logger.info(f'**** Could not charge. Check payment summary ****')
        logger.info(f'Payment summary: {summary}')
    except Exception as e:
        import traceback

        traceback.print_exc()
        # Do not continue with request if payment fails
        raise e


def get_summary():
    resp = payment_client.get_summary(token=authorized_jwt, app_id='search')  # type: ignore
    has_payment_method = resp['data'].get('hasPaymentMethod', False)
    user_credits = resp['data'].get('credits', None)
    internal_product_id = resp['data'].get('internalProductId', None)
    user_account_status = resp['data'].get('userAccountStatus', None)
    return {
        'has_payment_method': has_payment_method,
        'credits': user_credits,
        'internal_product_id': internal_product_id,
        'user_account_status': user_account_status,
    }


def can_charge(summary):
    return (
        (summary['credits'] is not None and summary['credits'] > 0)
        or summary['has_payment_method']
    ) and summary['user_account_status'] == 'active'



from typing import List

from pydantic import Field

from now.executor.gateway.bff.app.v1.models.shared import BaseRequestModel


class UpdateEmailsRequestModel(BaseRequestModel):
    user_emails: List[str] = Field(
        default=...,
        description='List of user emails who are allowed to access the flow',
    )


class UpdateApiKeysRequestModel(BaseRequestModel):
    api_keys: List[str] = Field(
        default=...,
        description='List of api keys which allow to access the flow in an automated way',
    )



from math import ceil, sqrt
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field

from now.executor.gateway.bff.app.v1.models.shared import (
    BaseRequestModel,
    ModalityModel,
)


class _NamedScore(BaseModel):
    value: Optional[float] = None


class IndexRequestModel(BaseRequestModel):
    data: List[Tuple[Dict[str, ModalityModel], Dict[str, Any]]] = Field(
        default=[({}, {})],
        description='List of tuples where each tuple contains a dictionary of data and a dictionary of tags. '
        'The data dictionary maps the field name to its value. ',
        example=[({'title': ModalityModel(text='this title')}, {'color': 'red'})],
    )


class SearchRequestModel(BaseRequestModel):
    limit: int = Field(
        default=10, description='Number of matching results to return', example=10
    )
    filters: Optional[Dict[str, Union[List, Dict[str, Union[int, float]]]]] = Field(
        default={},
        description='dictionary with filters for search results',
        example={'color': ['blue'], 'price': {'lt': 50.0}},
    )
    query: List[Dict] = Field(
        default={},
        description='List of dictionaries with query fields. Each dictionary represents a field in the query.',
        example=[
            {'name': 'title', 'modality': 'text', 'value': 'cute cats'},
            {
                'name': 'image',
                'modality': 'image',
                'value': 'https://example.com/image.jpg',
            },
        ],
    )
    create_temp_link: bool = Field(
        default=False,
        description='If true, a temporary link to the file is created. '
        'This is useful if the file is stored in a cloud bucket.',
        example=False,
    )
    score_calculation: List[List] = Field(
        default=[],
        description='List of lists, where each nested list contains a query_field, index_field, matching_method and weight.'
        ' This defines how scores should be calculated for documents. The matching_method can be an encoder name or '
        'bm25. The weight is a float which is used to scale the score.',
        example=[['query_text', 'title', 'encoderclip', 1.0]],
    )
    get_score_breakdown: bool = Field(
        default=False,
        description='If true, the score breakdown is returned in the response tags.',
        example=True,
    )


class SearchResponseModel(BaseModel):
    id: str = Field(
        default=...,
        nullable=False,
        description='Id of the matching result.',
        example='123',
    )
    scores: Optional[Dict[str, '_NamedScore']] = Field(
        description='Similarity score with respect to the query.',
        example={'score': {'value': 0.5}},
    )
    tags: Optional[
        Dict[
            str,
            Union[
                Optional[Union[str, bool, float]],
                List[Optional[Union[str, bool, float]]],
                Dict[str, Optional[Union[str, bool, float]]],
            ],
        ]
    ] = Field(
        description='Additional tags associated with the file.',
        example={'price': {'lt': 50.0}},
    )
    fields: Dict[str, ModalityModel] = Field(
        default={},
        description='Dictionary which maps the field name to its value.',
        example={
            'title': {'text': 'hello world'},
            'image': {'uri': 'https://example.com/image.jpg'},
        },
    )

    def __init__(
        self,
        id: str,
        scores: Optional[Dict[str, '_NamedScore']] = {},
        tags: Optional[
            Dict[
                str,
                Union[
                    Optional[Union[str, bool, int, float]],
                    List[Optional[Union[str, bool, int, float]]],
                    Dict[str, Optional[Union[str, bool, int, float]]],
                ],
            ]
        ] = {},
        fields: Dict[str, ModalityModel] = {},
    ) -> None:
        super().__init__(id=id, scores=scores, fields=fields)
        self.validate_tags(tags)
        self.tags = tags

    def validate_tags(self, tags):
        for key, value in tags.items():
            if isinstance(value, list):
                for item in value:
                    self.validate_tags({'': item})
            elif isinstance(value, dict):
                for _key, _value in value.items():
                    self.validate_tags({_key: _value})
            elif not isinstance(value, (str, bool, int, float)):
                raise ValueError(
                    f"Invalid type '{type(item)}' of value '{item}' for key '{key}' in tags"
                )

    def to_html(self, disable_to_datauri: bool = False) -> str:
        """Converts the SearchResponseModel to HTML. This is used to display the a single multi-modal result as HTML.

        :param disable_to_datauri: If True, the image is not converted to datauri.
        """
        # sort dictionary by keys, to have the same order in displaying elements
        single_fields_in_html = [
            mm.to_html(title, disable_to_datauri)
            for title, mm in dict(sorted(self.fields.items())).items()
        ]
        mm_in_html = ''.join(single_fields_in_html)
        return mm_in_html

    @classmethod
    def responses_to_html(
        cls, responses: List['SearchResponseModel'], disable_to_datauri: bool = False
    ) -> str:
        """Converts a list of SearchResponseModel to HTML. This is used to display the multi-modal results as HTML."""
        html_list = [r.to_html(disable_to_datauri) for r in responses]

        num_html = len(html_list)
        side_length = ceil(sqrt(num_html))
        output_html = "<div style='display: grid; grid-template-columns: repeat({0}, 1fr); grid-gap: 10px;'>".format(
            side_length
        )

        for i in range(num_html):
            output_html += (
                "<div style='border: 1px solid black; padding: 5px;'>{0}</div>".format(
                    html_list[i]
                )
            )

        output_html += "</div>"

        return output_html

    class Config:
        case_sensitive = False
        arbitrary_types_allowed = True


class SuggestionRequestModel(BaseRequestModel):
    text: Optional[str] = Field(default=None, description='Text', example='cute cats')


IndexRequestModel.update_forward_refs()
SearchRequestModel.update_forward_refs()
SearchResponseModel.update_forward_refs()



import yaml
from jina.jaml import JAML


def get_flow_id(host):
    return host[len('https://') : -len('-http.wolf.jina.ai')]


class Dumper(yaml.Dumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)


def write_flow_file(flow_yaml_content, new_yaml_file_path):
    with open(new_yaml_file_path, 'w') as f:
        JAML.dump(
            flow_yaml_content,
            f,
            indent=2,
            allow_unicode=True,
            Dumper=Dumper,
        )



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import importlib
import os
import uuid

from hubble import AuthenticationRequiredError

from now.common.detect_schema import (
    set_field_names_elasticsearch,
    set_field_names_from_docarray,
    set_field_names_from_local_folder,
    set_field_names_from_s3_bucket,
)
from now.constants import MODALITY_TO_MODELS, DatasetTypes
from now.deployment.deployment import cmd
from now.log import yaspin_extended
from now.now_dataclasses import DialogOptions, UserInput
from now.utils.authentication.helpers import (
    get_aws_profile,
    get_info_hubble,
    jina_auth_login,
)
from now.utils.common.helpers import hide_string_chars, sigmap, to_camel_case
from now.utils.errors.helpers import DemoAvailableException, RetryException

AVAILABLE_SOON = 'will be available in upcoming versions'
aws_profile = get_aws_profile()

# Make sure you add this dialog option to your app in order of dependency, i.e., if some dialog option depends on other
# than the parent should be called first before the dependant can called.


def _check_index_field(user_input: UserInput, **kwargs):
    if not user_input.index_fields:
        raise RetryException('Please select at least one index field')

    if '__all__' in user_input.index_fields:
        user_input.index_fields = list(
            user_input.index_field_candidates_to_modalities.keys()
        )
    elif any(
        idx_field
        for idx_field in user_input.index_fields
        if idx_field not in list(user_input.index_field_candidates_to_modalities.keys())
    ):
        raise ValueError(
            f'Index field specified {user_input.index_fields} is not among the index candidate fields. Please '
            f'choose one of the following: {list(user_input.index_field_candidates_to_modalities.keys())}'
        )


def _fill_filter_field_if_selected_all(user_input: UserInput, **kwargs):
    if '__all__' in user_input.filter_fields:
        user_input.filter_fields = list(
            user_input.filter_field_candidates_to_modalities.keys()
        )


def _append_all_option_to_filter(user_input: UserInput):
    choices = [
        {'name': field, 'value': field}
        for field in user_input.filter_field_candidates_to_modalities.keys()
        if field not in user_input.index_fields
    ]
    if len(choices) > 1:
        choices.append({'name': 'All of the above', 'value': '__all__'})
    return choices


APP_NAME = DialogOptions(
    name='flow_name',
    prompt_message='Choose a name for your application:',
    prompt_type='input',
    is_terminal_command=True,
    post_func=lambda user_input, **kwargs: clean_flow_name(user_input),
)


def clean_flow_name(user_input: UserInput):
    """
    Clean the flow name to make it valid, removing special characters and spaces.
    """
    user_input.flow_name = ''.join(
        [c for c in user_input.flow_name or '' if c.isalnum() or c == '-']
    ).lower()


DATASET_TYPE = DialogOptions(
    name='dataset_type',
    prompt_message='How do you want to provide input? (format: https://docarray.jina.ai/)',
    choices=[
        {'name': 'Demo dataset', 'value': DatasetTypes.DEMO},
        {
            'name': 'DocumentArray name (recommended)',
            'value': DatasetTypes.DOCARRAY,
        },
        {
            'name': 'Local folder',
            'value': DatasetTypes.PATH,
        },
        {
            'name': 'S3 bucket',
            'value': DatasetTypes.S3_BUCKET,
        },
        {
            'name': 'Elasticsearch',
            'value': DatasetTypes.ELASTICSEARCH,
        },
    ],
    prompt_type='list',
    is_terminal_command=True,
    post_func=lambda user_input, **kwargs: check_login_dataset(user_input),
)


def check_login_dataset(user_input: UserInput):
    if user_input.dataset_type == DatasetTypes.DOCARRAY and user_input.jwt is None:
        _jina_auth_login(user_input)


def _get_demo_data_choices(user_input: UserInput, **kwargs):
    all_demo_datasets = []
    for demo_datasets in user_input.app_instance.demo_datasets.values():
        all_demo_datasets.extend(demo_datasets)
    return [
        {'name': demo_data.display_name, 'value': demo_data.name}
        for demo_data in all_demo_datasets
    ]


def _check_if_demo_available(user_input: UserInput, **kwargs):
    # If the demo is available then do not continue with the dialog and break
    if user_input.app_instance.is_demo_available(user_input):
        raise DemoAvailableException(
            'Demo is available. Suspending the remaining dialog'
        )
    # We need login only if the deployment is needed else
    # for hosted example user should have access without login
    if user_input.jwt is None:
        _jina_auth_login(user_input)
    set_field_names_from_docarray(user_input)


DEMO_DATA = DialogOptions(
    name='dataset_name',
    prompt_message='What demo dataset do you want to use?',
    choices=lambda user_input, **kwargs: _get_demo_data_choices(user_input),
    prompt_type='list',
    depends_on=DATASET_TYPE,
    is_terminal_command=True,
    description='Select one of the available demo datasets',
    conditional_check=lambda user_input, **kwargs: user_input.dataset_type
    == DatasetTypes.DEMO,
    post_func=lambda user_input, **kwargs: _check_if_demo_available(user_input),
)

DOCARRAY_NAME = DialogOptions(
    name='dataset_name',
    prompt_message='Please enter your DocArray name:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.DOCARRAY,
    post_func=lambda user_input, **kwargs: set_field_names_from_docarray(user_input),
)

DATASET_PATH = DialogOptions(
    name='dataset_path',
    prompt_message='Please enter the path to the local folder:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    is_terminal_command=True,
    conditional_check=lambda user_input: user_input.dataset_type == DatasetTypes.PATH,
    post_func=lambda user_input, **kwargs: set_field_names_from_local_folder(
        user_input
    ),
)


def _fix_s3_uri(user_input: UserInput):
    if not user_input.dataset_path.endswith('/'):
        user_input.dataset_path += '/'


DATASET_PATH_S3 = DialogOptions(
    name='dataset_path',
    prompt_message='Please enter the S3 URI to the folder:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
    post_func=lambda user_input, **kwargs: _fix_s3_uri(user_input),
)

AWS_ACCESS_KEY_ID = DialogOptions(
    name='aws_access_key_id',
    prompt_message=f'Please enter the AWS access key ID: [{hide_string_chars(aws_profile.aws_access_key_id)}]',
    default=f'{aws_profile.aws_access_key_id}',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
    is_terminal_command=True,
)

AWS_SECRET_ACCESS_KEY = DialogOptions(
    name='aws_secret_access_key',
    prompt_message=f'Please enter the AWS secret access key: [{hide_string_chars(aws_profile.aws_secret_access_key)}]',
    default=f'{aws_profile.aws_secret_access_key}',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
    is_terminal_command=True,
)

AWS_REGION_NAME = DialogOptions(
    name='aws_region_name',
    prompt_message=f'Please enter the AWS region: [{aws_profile.region}]',
    default=f'{aws_profile.region}',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
    post_func=lambda user_input, **kwargs: set_field_names_from_s3_bucket(user_input),
    is_terminal_command=True,
)

# --------------------------------------------- #


INDEX_FIELDS = DialogOptions(
    name='index_fields',
    choices=lambda user_input, **kwargs: [
        {'name': field, 'value': field}
        for field in user_input.index_field_candidates_to_modalities.keys()
    ]
    + (
        [{'name': 'All of the above', 'value': '__all__'}]
        if len(user_input.index_field_candidates_to_modalities) > 1
        else []
    ),
    prompt_message='Please select the index fields:',
    prompt_type='checkbox',
    is_terminal_command=True,
    post_func=_check_index_field,
    argparse_kwargs={
        'type': lambda fields: fields.split(',') if fields else UserInput().index_fields
    },
)

FILTER_FIELDS = DialogOptions(
    name='filter_fields',
    choices=lambda user_input, **kwargs: _append_all_option_to_filter(user_input),
    prompt_message='Please select the filter fields',
    prompt_type='checkbox',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.filter_field_candidates_to_modalities
    is not None
    and len(
        set(user_input.filter_field_candidates_to_modalities.keys())
        - set(user_input.index_fields)
    )
    > 0,
    post_func=_fill_filter_field_if_selected_all,
)


def update_model_choice(user_input: UserInput, option_name, **kwargs):
    if not kwargs.get(option_name):
        raise RetryException('Please select at least one model')

    user_input.model_choices[option_name] = kwargs.get(option_name)


def get_models_dialog(user_input: UserInput):
    models_dialog = []
    for index_field in user_input.index_fields:
        option_name = f'{index_field}_model'
        models_dialog.append(
            DialogOptions(
                name=option_name,
                prompt_message=f'Please select the model for {index_field}:',
                prompt_type='checkbox',
                choices=MODALITY_TO_MODELS[
                    user_input.index_field_candidates_to_modalities[index_field]
                ],
                post_func=lambda user_inp, opt_name=option_name, **kwargs: update_model_choice(
                    user_inp, opt_name, **kwargs
                ),
            )
        )

    return models_dialog


MODEL_SELECTION = DialogOptions(
    name='model_selection',
    prompt_message='This is a dynamic dialog which will spawn multiple dialog options',
    prompt_type='list',
    choices=[],
    depends_on=INDEX_FIELDS,
    conditional_check=lambda user_input: user_input.index_fields,
    is_terminal_command=True,
    dynamic_func=get_models_dialog,
)

ES_INDEX_NAME = DialogOptions(
    name='es_index_name',
    prompt_message='Please enter the name of your Elasticsearch index:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

ES_HOST_NAME = DialogOptions(
    name='es_host_name',
    prompt_message='Please enter the address of your Elasticsearch node:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

ES_ADDITIONAL_ARGS = DialogOptions(
    name='es_additional_args',
    prompt_message='Please enter additional arguments for your Elasticsearch node if there are any:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.ELASTICSEARCH,
    post_func=lambda user_input, **kwargs: set_field_names_elasticsearch(user_input),
)

# --------------------------------------------- #
SECURED = DialogOptions(
    name='secured',
    prompt_message='Do you want to secure the Flow?',
    prompt_type='list',
    choices=[
        {'name': '⛔ no', 'value': False},
        {'name': '✅ yes', 'value': True},
    ],
    is_terminal_command=True,
    post_func=lambda user_input, **kwargs: check_login_deployment(user_input),
)


def check_login_deployment(user_input: UserInput):
    if user_input.jwt is None:
        _jina_auth_login(user_input)


API_KEY = DialogOptions(
    name='api_key',
    prompt_message='Do you want to generate an API key to access this deployment?',
    prompt_type='list',
    choices=[
        {'name': '✅ yes', 'value': uuid.uuid4().hex},
        {'name': '⛔ no', 'value': False},
    ],
    depends_on=SECURED,
    is_terminal_command=True,
    description='Pass an API key to access the Flow once the deployment is complete. ',
    conditional_check=lambda user_inp: str(user_inp.secured).lower() == 'true',
    post_func=lambda user_input, **kwargs: _set_value_to_none(user_input),
)

ADDITIONAL_USERS = DialogOptions(
    name='additional_user',
    prompt_message='Do you want to provide additional users access to this flow?',
    prompt_type='list',
    choices=[
        {'name': '✅ yes', 'value': True},
        {'name': '⛔ no', 'value': False},
    ],
    depends_on=SECURED,
    conditional_check=lambda user_inp: str(user_inp.secured).lower() == 'true',
)

USER_EMAILS = DialogOptions(
    name='user_emails',
    prompt_message='Please enter email addresses (separated by commas) '
    'to grant access to this Flow.\nAdditionally, you can specify comma-separated domain names'
    ' such that all users from that domain can access this Flow, e.g. `jina.ai`\n',
    prompt_type='input',
    depends_on=ADDITIONAL_USERS,
    conditional_check=lambda user_inp: user_inp.additional_user,
    post_func=lambda user_input, **kwargs: _add_additional_users(user_input, **kwargs),
)


def _set_value_to_none(user_input: UserInput):
    if not user_input.api_key:
        user_input.api_key = None


def _add_additional_users(user_input: UserInput, **kwargs):
    if not kwargs.get('user_emails', None):
        raise RetryException('Please provide at least one email address')
    user_input.user_emails = (
        [email.strip() for email in kwargs['user_emails'].split(',')]
        if kwargs['user_emails']
        else []
    )


def construct_app(app_name: str):
    return getattr(
        importlib.import_module(f'now.app.{app_name}.app'),
        f'{to_camel_case(app_name)}',
    )()


def _jina_auth_login(user_input: UserInput, **kwargs):
    try:
        jina_auth_login()
    except AuthenticationRequiredError:
        with yaspin_extended(
            sigmap=sigmap, text='Log in to Jina AI Cloud', color='green'
        ) as spinner:
            cmd('jina auth login')
        spinner.ok('🛠️')

    get_info_hubble(user_input)
    os.environ['JCLOUD_NO_SURVEY'] = '1'


app_config = [APP_NAME]
data_type = [DATASET_TYPE]
data_fields = [INDEX_FIELDS, FILTER_FIELDS]
get_models = [MODEL_SELECTION]
data_demo = [DEMO_DATA]
data_da = [DOCARRAY_NAME, DATASET_PATH]
data_s3 = [DATASET_PATH_S3, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_REGION_NAME]
data_es = [
    ES_HOST_NAME,
    ES_INDEX_NAME,
    ES_ADDITIONAL_ARGS,
]
remote_cluster = [SECURED, API_KEY, ADDITIONAL_USERS, USER_EMAILS]

base_options = (
    data_type
    + data_demo
    + data_da
    + data_s3
    + data_es
    + data_fields
    + get_models
    + app_config
    + remote_cluster
)



from fastapi import APIRouter

from now.executor.gateway.bff.app.v1.routers.admin import router as admin_router
from now.executor.gateway.bff.app.v1.routers.info import router as info_router
from now.executor.gateway.bff.app.v1.routers.search import router as search_router

search_app_router = APIRouter()
search_app_router.include_router(search_router, tags=['Search App'])
search_app_router.include_router(info_router, tags=['Info'])


# rename admin_router to something more meaningful
extras_router = APIRouter()
extras_router.include_router(admin_router, tags=['Admin level operations'])



import os

import hubble


def get_default_request_body(secured):
    request_body = {}
    if secured:
        if 'WOLF_TOKEN' in os.environ:
            os.environ['JINA_AUTH_TOKEN'] = os.environ['WOLF_TOKEN']
        request_body['jwt'] = {'token': hubble.get_token()}
    return request_body



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import inspect
import pathlib
from typing import Dict, List, Optional, Union

from pyfiglet import Figlet

from now.common import options
from now.common.options import construct_app
from now.constants import MODALITY_TO_MODELS, Apps, DialogStatus
from now.now_dataclasses import DialogOptions, UserInput
from now.thirdparty.PyInquirer.prompt import prompt
from now.utils.errors.helpers import DemoAvailableException, RetryException

cur_dir = pathlib.Path(__file__).parent.resolve()


def configure_user_input(**kwargs) -> UserInput:
    user_input = UserInput()
    print_headline()
    # Create the search app.
    # TODO: refactor this when more apps are added
    user_input.app_instance = construct_app(Apps.SEARCH_APP)
    # Ask the options
    for option in options.base_options + user_input.app_instance.options:
        if configure_option(option, user_input, **kwargs) == DialogStatus.BREAK:
            break

    return user_input


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search use case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina NOW trains a model, pushes it to Jina AI Cloud '
        'and deploys a Flow and playground app in the cloud or locally. \nCheck out our demos or bring '
        'your own data.\n'
    )
    print('Visit docs.jina.ai to learn more about our framework')
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def configure_option(
    option: DialogOptions, user_input: UserInput, **kwargs
) -> DialogStatus:
    # Check if it is dynamic. If it is then spawn multiple dialogs for each option and return with continue
    if option.dynamic_func:
        if option.name in kwargs:
            # Expand dynamic options from parent option, expect a dict (supports only model_selection)
            expand_options_from_parent(kwargs, option, user_input)

        for result in option.dynamic_func(user_input):
            configure_option(result, user_input, **kwargs)
        return DialogStatus.CONTINUE

    # Check if it is dependent on some other dialog options
    if option.depends_on and not option.conditional_check(user_input):
        return DialogStatus.SKIP

    # Populate choices if needed
    if option.choices and inspect.isfunction(option.choices):
        option.choices = option.choices(user_input, **kwargs)

    while True:
        val = prompt_value(
            **option.__dict__,
            **kwargs,
        )

        if val:
            kwargs[option.name] = val
            if hasattr(user_input, option.name):
                setattr(user_input, option.name, val)

        try:
            # If there is any post function then invoke that
            if inspect.isfunction(option.post_func):
                option.post_func(user_input, **kwargs)
        except RetryException as e:
            print(e)
            continue
        except DemoAvailableException:
            return DialogStatus.BREAK
        break

    return DialogStatus.CONTINUE


def expand_options_from_parent(kwargs, option, user_input):
    for user_selection in kwargs[option.name].split(","):
        if ":" in user_selection:
            option_name, option_values = user_selection.split(":")
            kwargs[f"{option_name}_model"] = []
            if not option_name in user_input.index_field_candidates_to_modalities:
                raise ValueError(
                    f"Error with --{option.name}: `{option_name}` is not an index field."
                )
            for option_value in option_values.split("+"):
                model_selection = [
                    model
                    for model in MODALITY_TO_MODELS[
                        user_input.index_field_candidates_to_modalities[option_name]
                    ]
                    if model["name"] == option_value
                ]
                if model_selection:
                    kwargs[f"{option_name}_model"].append(model_selection[0]["value"])
                else:
                    model_choices = [
                        model["name"]
                        for model in MODALITY_TO_MODELS[
                            user_input.index_field_candidates_to_modalities[option_name]
                        ]
                    ]
                    raise ValueError(
                        f"Error with --{option.name}: `{option_value}` is not available. "
                        f"for index field `{option_name}`. Choices are: {','.join(model_choices)}."
                    )


def prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
    return maybe_prompt_user(qs, name, **kwargs)


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and attribute in kwargs:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        return answer[attribute]



import itertools
import json
import os
from typing import Optional

from better_profanity import profanity
from docarray import DocumentArray
from fast_autocomplete import AutoComplete

from now.executor.abstract.auth.auth import (
    SecurityLevel,
    get_auth_executor_class,
    secure_request,
)

Executor = get_auth_executor_class()


class NOWAutoCompleteExecutor2(Executor):
    def __init__(self, words=None, *args, **kwargs):
        self.words = words if words else {}
        self.auto_complete = None
        self.char_threshold = 100

        super().__init__(*args, **kwargs)

        self.words_path = (
            os.path.join(self.workspace, 'words.json') if self.workspace else None
        )
        if self.words_path and os.path.exists(self.words_path):
            with open(self.words_path, 'r') as fp:
                self.words = json.load(fp)
                self.auto_complete = AutoComplete(words=self.words)

    def update_save_words(self, word):
        """
        Method to update word count dictionary and dumps new dictionary.
        """
        if word in self.words:
            self.words[word]['count'] += 1
        else:
            self.words[word] = {'count': 1}
        if self.words_path:
            with open(self.words_path, 'w') as fp:
                json.dump(self.words, fp)

    @secure_request(on='/search', level=SecurityLevel.USER)
    def search_update(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        for doc in docs['@c']:
            if doc.text and not profanity.contains_profanity(doc.text):
                search_words = doc.text.split(' ')
                # prevent users from misusing API
                if len(doc.text) < self.char_threshold:
                    # in case query is composed of two words
                    for word in search_words:
                        self.update_save_words(word)
                    # add bi-gram and tri-gram suggestions
                    if len(search_words) == 2 or len(search_words) == 3:
                        self.update_save_words(doc.text)

        self.auto_complete = AutoComplete(words=self.words)
        return docs

    @secure_request(on='/suggestion', level=SecurityLevel.USER)
    def get_suggestion(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        for doc in docs:
            doc.tags['suggestions'] = self.flatten_list(
                self.auto_complete.search(doc.text, max_cost=3, size=5)
            )
        return docs

    def flatten_list(self, regular_list):
        flat_list = list(itertools.chain(*regular_list))
        return flat_list



from typing import TypeVar

import docarray


def docarray_typing_to_modality_string(T: TypeVar) -> str:
    """E.g. docarray.typing.Image -> image"""
    return T.__name__.lower()


def modality_string_to_docarray_typing(s: str) -> TypeVar:
    """E.g. image -> docarray.typing.Image"""
    return getattr(docarray.typing, s.capitalize())


def get_chunk_by_field_name(doc, field_name):
    """
    Gets a specific chunk by field name, using its position instead of getting the attribute directly.
    This solves the getattr problem when there are conflicting attributes with Document.
    :param doc: Document to get the chunk from.
    :param field_name: Field needed to extract the position.
    :return: Specific chunk by field.
    """
    try:
        field_position = int(
            doc._metadata['multi_modal_schema'][field_name]['position']
        )
        return doc.chunks[field_position]
    except Exception as e:
        raise e



import os
from typing import Dict, List, Tuple, TypeVar

from jina import Client

from now.app.base.app import JinaNOWApp
from now.constants import (
    ACCESS_PATHS,
    DEMO_NS,
    EXTERNAL_CLIP_HOST,
    EXTERNAL_SBERT_HOST,
    NOW_AUTOCOMPLETE_VERSION,
    NOW_ELASTIC_INDEXER_VERSION,
    NOW_PREPROCESSOR_VERSION,
    Apps,
    DatasetTypes,
    Models,
)
from now.demo_data import AVAILABLE_DATASETS, DemoDataset, DemoDatasetNames
from now.executor.name_to_id_map import name_to_id_map
from now.now_dataclasses import UserInput


class SearchApp(JinaNOWApp):
    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.SEARCH_APP

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Search app'

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 8

    @property
    def demo_datasets(self) -> Dict[TypeVar, List[DemoDataset]]:
        return AVAILABLE_DATASETS

    @property
    def finetune_datasets(self) -> [Tuple]:
        return DemoDatasetNames.DEEP_FASHION, DemoDatasetNames.BIRD_SPECIES

    def is_demo_available(self, user_input) -> bool:
        if (
            user_input.dataset_type == DatasetTypes.DEMO
            and 'NOW_EXAMPLES' not in os.environ
            and 'NOW_CI_RUN' not in os.environ
        ):
            client = Client(
                host=f'grpcs://{DEMO_NS.format(user_input.dataset_name.split("/")[-1])}.dev.jina.ai'
            )
            try:
                client.post('/dry_run')
            except Exception as e:  # noqa E722
                pass
            return True
        return False

    @staticmethod
    def autocomplete_stub(testing=False) -> Dict:
        return {
            'name': 'autocomplete_executor',
            'uses': f'jinahub+docker://{name_to_id_map.get("NOWAutoCompleteExecutor2")}/{NOW_AUTOCOMPLETE_VERSION}'
            if not testing
            else 'NOWAutoCompleteExecutor2',
            'needs': 'gateway',
            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
            'jcloud': {
                'autoscale': {
                    'min': 0,
                    'max': 1,
                    'metric': 'concurrency',
                    'target': 1,
                },
                'resources': {
                    'instance': 'C1',
                    'capacity': 'spot',
                    'storage': {'kind': 'efs', 'size': '1M'},
                },
            },
        }

    @staticmethod
    def preprocessor_stub(testing=False) -> Dict:
        return {
            'name': 'preprocessor',
            'needs': 'autocomplete_executor',
            'uses': f'jinahub+docker://{name_to_id_map.get("NOWPreprocessor")}/{NOW_PREPROCESSOR_VERSION}'
            if not testing
            else 'NOWPreprocessor',
            'jcloud': {
                'autoscale': {
                    'min': 0,
                    'max': 100,
                    'metric': 'concurrency',
                    'target': 1,
                },
                'resources': {'instance': 'C4', 'capacity': 'spot'},
            },
            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
        }

    @staticmethod
    def clip_encoder_stub() -> Tuple[Dict, int]:
        return {
            'name': Models.CLIP_MODEL,
            'uses': f'jinahub+docker://CLIPOnnxEncoder/0.8.1-gpu',
            'host': EXTERNAL_CLIP_HOST,
            'port': 443,
            'tls': True,
            'external': True,
            'uses_with': {'access_paths': ACCESS_PATHS, 'name': 'ViT-B-32::openai'},
            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
            'needs': 'preprocessor',
        }, 512

    @staticmethod
    def sbert_encoder_stub() -> Tuple[Dict, int]:
        return {
            'name': Models.SBERT_MODEL,
            'uses': f'jinahub+docker://TransformerSentenceEncoder',
            'uses_with': {
                'access_paths': ACCESS_PATHS,
                'model_name': 'msmarco-distilbert-base-v3',
            },
            'host': EXTERNAL_SBERT_HOST,
            'port': 443,
            'tls': True,
            'external': True,
            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
            'needs': 'preprocessor',
        }, 768

    @staticmethod
    def indexer_stub(
        user_input: UserInput,
        encoder2dim: Dict[str, int],
        testing=False,
    ) -> Dict:
        """Creates indexer stub.

        :param user_input: user input
        :param encoder2dim: maps encoder name to its output dimension
        :param testing: use local executors if True
        """
        document_mappings_list = []

        for encoder, dim in encoder2dim.items():
            document_mappings_list.append(
                [
                    encoder,
                    dim,
                    [
                        user_input.field_names_to_dataclass_fields[
                            index_field.replace('_model', '')
                        ]
                        for index_field, encoders in user_input.model_choices.items()
                        if encoder in encoders
                    ],
                ]
            )
        provision_index = 'yes' if not testing else 'no'
        provision_shards = os.getenv('PROVISION_SHARDS', '1')
        provision_replicas = os.getenv('PROVISION_REPLICAS', '0')

        return {
            'name': 'indexer',
            'needs': list(encoder2dim.keys()),
            'uses': f'jinahub+docker://{name_to_id_map.get("NOWElasticIndexer")}/{NOW_ELASTIC_INDEXER_VERSION}'
            if not testing
            else 'NOWElasticIndexer',
            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
            'uses_with': {
                'document_mappings': document_mappings_list,
            },
            'no_reduce': True,
            'jcloud': {
                'labels': {
                    'app': 'indexer',
                    'provision-index': provision_index,
                    'provision-shards': provision_shards,
                    'provision-replicas': provision_replicas,
                },
                'resources': {'instance': 'C2', 'capacity': 'spot'},
            },
        }

    def get_executor_stubs(
        self, user_input: UserInput, testing=False, **kwargs
    ) -> List[Dict]:
        """Returns a dictionary of executors to be added in the flow

        :param user_input: user input
        :param testing: use local executors if True
        :return: executors stubs with filled-in env vars
        """
        flow_yaml_executors = [
            self.autocomplete_stub(testing),
            self.preprocessor_stub(testing),
        ]

        encoder2dim = {}

        def add_encoders_to_flow(models):
            for model, encoder_stub in models:
                if any(
                    model in user_input.model_choices[f"{field}_model"]
                    for field in user_input.index_fields
                ):
                    encoder, dim = encoder_stub()
                    encoder2dim[encoder['name']] = dim
                    flow_yaml_executors.append(encoder)

        add_encoders_to_flow(
            [
                (Models.SBERT_MODEL, self.sbert_encoder_stub),
                (Models.CLIP_MODEL, self.clip_encoder_stub),
            ]
        )

        flow_yaml_executors.append(
            self.indexer_stub(
                user_input,
                encoder2dim=encoder2dim,
                testing=testing,
            )
        )

        return flow_yaml_executors

    @property
    def max_request_size(self) -> int:
        """Max number of documents in one request"""
        return 10



import json
import os
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from time import sleep
from typing import Dict, List, Tuple

import pandas as pd
import requests
from docarray import Document, DocumentArray
from tqdm import tqdm

from now.admin.utils import get_default_request_body
from now.executor.gateway.bff.app.v1.models.search import SearchResponseModel
from now.utils.docarray.helpers import get_chunk_by_field_name


def compare_flows_for_queries(
    da: DocumentArray,
    flow_ids_http_score_calculation: List[Tuple],
    limit: int,
    results_per_table: int = 20,
    disable_to_datauri: bool = False,
):
    """This function compares flows by querying every flow with the multi-modal query documents of da and retrieving
    limit results from them. It converts the responses into HTML which are then saved as a table to disk. It splits
    the created HTML into multiple files if there are a lot of queries.

    :param da: DocumentArray of queries defined as multi-modal documents
    :param flow_ids_http_score_calculation: a list consisting of tuples which are (flow ID, host of HTTP gateway,
    score calculation)
    :param limit: the number of results to retrieve
    :param results_per_table: number of queries which should be put into the same table
    :param disable_to_datauri: if True, the images are not converted to DataURI
    """
    # call flow/api/v1/search-app/search for all queries and all flow_ids_http_score_calculation
    print(f'Comparing {len(da)} queries')
    now = datetime.now()
    folder = str(
        os.path.abspath(
            f'compare'
            f'-{"-".join(set([cihs[0] for cihs in flow_ids_http_score_calculation]))}'
            f'-limit_{limit}'
            f'-{now.strftime("%Y%m%d")}-{now.strftime("%H%M")}'
        )
    )
    os.mkdir(folder)
    rows = []
    cnt_tables = 0
    with tqdm(total=len(da)) as pbar:
        with ProcessPoolExecutor(max_workers=min(len(da), 20)) as ex:
            futures = [
                ex.submit(
                    _evaluate_query,
                    query,
                    flow_ids_http_score_calculation,
                    limit,
                    disable_to_datauri,
                )
                for query in da
            ]
            for future in futures:
                rows.append(future.result())
                pbar.update(1)
                if len(rows) == results_per_table:
                    df = pd.DataFrame(rows)
                    df.to_html(
                        os.path.join(
                            folder,
                            f'queries-{cnt_tables * results_per_table}-to-{(cnt_tables + 1) * results_per_table}.html',
                        ),
                        escape=False,
                    )
                    cnt_tables += 1
                    rows = []
    if rows:
        df = pd.DataFrame(rows[cnt_tables : len(rows)])
        df.to_html(
            os.path.join(
                folder,
                f'queries-{cnt_tables * results_per_table}-to-{cnt_tables * results_per_table + len(rows) - 1}.html',
            ),
            escape=False,
        )
    print(f'Comparison tables were saved as HTMLs under: {folder}')


def _evaluate_query(
    query: Document,
    flow_ids_http_score_calculation: List[Tuple],
    limit: int,
    disable_to_datauri: bool,
) -> Dict[str, str]:
    """Sends the query to each flow and returns a dictionary which maps 'query' and flow names with their semantic
    scores to HTML of the results.

    :param query: query defined as multi-modal document
    :param flow_ids_http_score_calculation: a list consisting of tuples which are (flow ID, host of HTTP gateway,
    score calculation)
    :param limit: the number of results to retrieve
    :param disable_to_datauri: if True, the images are not converted to DataURI
    """
    query_dict_search_request = []
    query_dict_response = {}
    for field in query._metadata['multi_modal_schema'].keys():
        field_chunk = get_chunk_by_field_name(query, field)
        if field_chunk.modality != 'text':
            if field_chunk.uri:
                field_chunk.convert_uri_to_datauri()
            else:
                field_chunk.convert_content_to_datauri()
        query_dict_response[field] = {
            'uri': field_chunk.uri,
            'blob': field_chunk.blob,
            'text': field_chunk.text,
        }
        query_dict_search_request.append(
            {
                'name': field,
                'modality': field_chunk.modality,
                'value': field_chunk.uri or field_chunk.content,
            }
        )
    row = {
        'query': SearchResponseModel(id=query.id, fields=query_dict_response).to_html(
            disable_to_datauri
        )
    }

    for flow_name, http_host, score_calculation in flow_ids_http_score_calculation:
        request_body = get_default_request_body(secured=True)
        request_body['limit'] = limit
        request_body['query'] = query_dict_search_request
        request_body['create_temp_link'] = True
        request_body['score_calculation'] = score_calculation
        for _ in range(5):
            try:
                response = requests.post(
                    f'{http_host}/api/v1/search-app/search', json=request_body
                )
                if response.status_code == 200:
                    break
            except ConnectionError:
                sleep(1)
                continue
        row[
            f'{flow_name} - {json.dumps(score_calculation)}'
        ] = SearchResponseModel.responses_to_html(
            [SearchResponseModel(**r) for r in response.json()], disable_to_datauri
        )

    return row



import base64
import itertools
import os
from copy import deepcopy
from tempfile import TemporaryDirectory
from typing import Dict, Type, Union

import filetype
from docarray import Document, DocumentArray
from docarray.typing import Text
from fastapi import HTTPException, status
from jina.serve.runtimes.gateway.streamer import GatewayStreamer

from now.constants import SUPPORTED_FILE_TYPES


def field_dict_to_mm_doc(
    field_dict: Dict,
    data_class: Type,
    modalities_dict: Dict,
    field_names_to_dataclass_fields={},
) -> Document:
    """Converts a dictionary of field names to their values to a document.
    :param field_dict: key-value pairs of field names and their values
    :param data_class: @docarray.dataclass class which encapsulates the fields of the multimodal document
    :param modalities_dict: dictionary of field names to their modalities
    :param field_names_to_dataclass_fields: mapping of field names to data class fields (e.g. {'title': 'text_0'})
    :return: multi-modal document
    """

    with TemporaryDirectory() as tmp_dir:
        try:
            if field_names_to_dataclass_fields:
                field_dict_orig = deepcopy(field_dict)
                modalities_dict_orig = deepcopy(modalities_dict)
                field_dict = {
                    field_name_data_class: field_dict_orig[file_name]
                    for file_name, field_name_data_class in field_names_to_dataclass_fields.items()
                }
                modalities_dict = {
                    field_name_data_class: modalities_dict_orig[file_name]
                    for file_name, field_name_data_class in field_names_to_dataclass_fields.items()
                }
            data_class_kwargs = {}
            for field_name_data_class, field_value in field_dict.items():
                # save blob into a temporary file such that it can be loaded by the multimodal class
                if modalities_dict[
                    field_name_data_class
                ] != Text and not field_value.startswith('http'):
                    base64_decoded = base64.b64decode(field_value.encode('utf-8'))
                    file_ending = filetype.guess(base64_decoded)
                    if not file_ending:
                        raise ValueError(
                            f'Could not guess file type of blob {field_value}. '
                            f'Please provide a valid file type.'
                        )
                    file_ending = file_ending.extension
                    if file_ending not in itertools.chain(
                        *SUPPORTED_FILE_TYPES.values()
                    ):
                        raise ValueError(
                            f'File type {file_ending} is not supported. '
                            f'Please provide a valid file type.'
                        )
                    file_path = os.path.join(
                        tmp_dir, field_name_data_class + '.' + file_ending
                    )
                    with open(file_path, 'wb') as f:
                        f.write(base64_decoded)
                    field_value = file_path
                if field_value:
                    data_class_kwargs[field_name_data_class] = field_value
                else:
                    raise ValueError(
                        f'Content of field {field_name_data_class} is None. '
                    )
            doc = Document(data_class(**data_class_kwargs))
        except BaseException as e:
            import traceback

            traceback.format_exc()
            raise HTTPException(
                status_code=500,
                detail=f'Not a correctly encoded request. Please see the error stack for more information. \n{traceback.format_exc()}',
            )

    return doc


async def jina_client_post(
    request_model,
    endpoint: str,
    docs: Union[Document, DocumentArray],
    parameters=None,
    *args,
    **kwargs,
) -> DocumentArray:
    """Posts to the endpoint of the Jina client.
    :param request_model: contains the request model of the flow
    :param endpoint: endpoint which shall be called, e.g. '/index' or '/search'
    :param docs: document(s) which shall be passed in
    :param parameters: parameters to pass to the executors, e.g. jwt for securitization or limit for search
    :param args: any additional arguments passed to the `client.post` method
    :param kwargs: any additional keyword arguments passed to the `client.post` method
    :return: response of `client.post`
    """
    if not isinstance(docs, DocumentArray):
        docs = DocumentArray([docs])
    streamer = GatewayStreamer.get_streamer()
    if parameters is None:
        parameters = {}
    auth_dict = {}
    if request_model.api_key is not None:
        auth_dict['api_key'] = request_model.api_key
    if request_model.jwt is not None:
        auth_dict['jwt'] = request_model.jwt

    async for response in streamer.stream_docs(
        exec_endpoint=endpoint,
        docs=docs,
        parameters={
            **auth_dict,
            **parameters,
            'access_paths': '@cc',
        },
        return_results=True,
        *args,
        **kwargs,
    ):
        if response.header.status.code == 1:
            raise_exception(
                name=response.header.status.exception.name,
                stacktrace=response.header.status.exception.stacks,
            )
        else:
            return response.docs


def raise_exception(
    name: str,
    stacktrace: str,
):
    if name == 'PermissionError':
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail='You are not authorised to use this flow',
        )
    else:
        raise HTTPException(
            status_code=500,
            detail=f'Request failed. Please see the error stack for more information. \n{stacktrace}',
        )



# -*- coding: utf-8 -*-
"""
`input` type question
"""
import inspect

from prompt_toolkit.lexers import SimpleLexer
from prompt_toolkit.shortcuts import prompt
from prompt_toolkit.validation import Validator, ValidationError

from .common import default_style

# use std prompt-toolkit control


def question(message, **kwargs):
    default = kwargs.pop('default', '')
    validate_prompt = kwargs.pop('validate', None)
    if validate_prompt:
        if inspect.isclass(validate_prompt) and issubclass(validate_prompt, Validator):
            kwargs['validator'] = validate_prompt()
        elif callable(validate_prompt):

            class _InputValidator(Validator):
                def validate(self, document):
                    verdict = validate_prompt(document.text)
                    if not verdict == True:
                        if verdict == False:
                            verdict = 'invalid input'
                        raise ValidationError(
                            message=verdict, cursor_position=len(document.text)
                        )

            kwargs['validator'] = _InputValidator()
    # TODO style defaults on detail level
    kwargs['style'] = kwargs.pop('style', default_style)
    qmark = kwargs.pop('qmark', '?')

    def _get_prompt_tokens():
        return [('class:questionmark', qmark), ('class:question', ' %s  ' % message)]

    return prompt(
        message=_get_prompt_tokens,
        lexer=SimpleLexer('class:answer'),
        default=default,
        **kwargs
    )



import base64
import logging
import os
from typing import Any, Dict, List

from fastapi import APIRouter, Body

from now.data_loading.create_dataclass import create_dataclass
from now.executor.gateway.bff.app.settings import user_input_in_bff
from now.executor.gateway.bff.app.v1.models.search import (
    SearchRequestModel,
    SearchResponseModel,
)
from now.executor.gateway.bff.app.v1.routers.helper import (
    field_dict_to_mm_doc,
    jina_client_post,
)
from now.executor.gateway.hubble_report import report_search_usage
from now.utils.docarray.helpers import (
    get_chunk_by_field_name,
    modality_string_to_docarray_typing,
)

logger = logging.getLogger(__file__)
logger.setLevel(os.environ.get('JINA_LOG_LEVEL', 'INFO'))

search_examples = {
    'working_text': {
        'summary': 'A working example: search with text',
        'description': 'A working example which can be tried out. Search with text on the best artworks dataset.',
        'value': {
            'limit': 10,
            'query': [
                {
                    'name': 'query_text_0',
                    'modality': 'text',
                    'value': 'cute cats',
                }
            ],
            'create_temp_link': False,
            'get_score_breakdown': True,
        },
    },
    'working_text_image': {
        'summary': 'A working example: search with text and image',
        'description': 'A working example which can be tried out. Search with text and image on the best artworks dataset.',
        'value': {
            'limit': 10,
            'query': [
                {
                    'name': 'query_text_0',
                    'modality': 'text',
                    'value': 'cute cats',
                },
                {
                    'name': 'query_image_0',
                    'modality': 'image',
                    'value': 'https://images.pexels.com/photos/45201/kitty-cat-kitten-pet-45201.jpeg',
                },
            ],
            'create_temp_link': False,
            'get_score_breakdown': True,
        },
    },
    'dummy': {
        'summary': 'A dummy example',
        'description': 'A dummy example,  do not run. For parameter reference only.',
        'value': {
            'limit': 10,
            'filters': {
                'color': ['blue', 'red'],
                'price': {'lte': 100, 'gte': 50},
            },
            'query': [
                {
                    'name': 'query_text_0',
                    'modality': 'text',
                    'value': 'cute cats',
                }
            ],
            'create_temp_link': False,
            'score_calculation': [['query_text_0', 'title', 'encoderclip', 1.0]],
        },
    },
}


router = APIRouter()


@router.post(
    '/search',
    response_model=List[SearchResponseModel],
    summary='Search data via query',
)
async def search(
    data: SearchRequestModel = Body(examples=search_examples),
):
    logger.info(f'Got search request: {data}')
    fields_modalities_mapping = {}
    fields_values_mapping = {}
    if len(data.query) == 0:
        raise ValueError('Query cannot be empty')

    for field in data.query:
        fields_modalities_mapping[field['name']] = modality_string_to_docarray_typing(
            field['modality']
        )
        fields_values_mapping[field['name']] = field['value']
    data_class, field_names_to_dataclass_fields = create_dataclass(
        fields=list(fields_modalities_mapping.keys()),
        fields_modalities=fields_modalities_mapping,
    )
    query_doc = field_dict_to_mm_doc(
        fields_values_mapping,
        data_class=data_class,
        modalities_dict=fields_modalities_mapping,
        field_names_to_dataclass_fields=field_names_to_dataclass_fields,
    )
    score_calculation = get_score_calculation(data, field_names_to_dataclass_fields)

    query_filter = {}
    for key, value in data.filters.items():
        key = 'tags__' + key
        query_filter[key] = value

    docs = await jina_client_post(
        endpoint='/search',
        docs=query_doc,
        parameters={
            'limit': data.limit,
            'filter': query_filter,
            'create_temp_link': data.create_temp_link,
            'score_calculation': score_calculation,
            'get_score_breakdown': data.get_score_breakdown,
        },
        request_model=data,
    )
    matches = []
    for doc in docs[0].matches:
        # todo: use multimodal doc in the future!
        scores = {}
        for score_name, named_score in doc.scores.items():
            scores[score_name] = named_score.to_dict()
        field_names_and_chunks = [
            [field_name, get_chunk_by_field_name(doc, field_name)]
            for field_name in doc._metadata['multi_modal_schema'].keys()
        ]
        results = {}
        for field_name, chunk in field_names_and_chunks:
            if chunk.blob:
                result = {'blob': base64.b64encode(chunk.blob).decode('utf-8')}
            elif chunk.text:
                result = {'text': chunk.text}
            elif chunk.uri:
                # in case we have content and uri, the content is preferred
                result = {'uri': chunk.uri}
            else:
                # We should not raise exception else it breaks the playground if a single chunk has no content
                # irrespective of what other chunks hold. We should just log it and move on.
                logger.info('Result without content', doc.id, doc.tags)
                result = {'text': ''}
            results[field_name] = result
        match = SearchResponseModel(
            id=doc.id,
            scores=scores,
            tags=doc.tags,
            fields=results,
        )
        matches.append(match)
    # reporting the usage at the end to make sure the request was successful

    logger.info(
        f'Reporting search usage after successful search request for user {data.jwt.get("token")}'
    )
    report_search_usage(user_token=data.jwt.get('token'))
    return matches


def get_score_calculation(
    data: SearchRequestModel, field_names_to_dataclass_fields: Dict[str, str]
) -> List[List[Any]]:
    """
    Extract and process the score calculation from the request model to the format expected by the indexer.
    This includes converting the field names to the dataclass field names, for the query and for the index fields.

    :param data: the request model
    :param field_names_to_dataclass_fields: a mapping from the field names in the request model to the field names in the dataclass
    :return: the score calculation in the format expected by the indexer. Example:
        [['query_text', 'my_product_image', 'encoderclip', 1], ['query_text', 'my_product_text', 'bm25', 1]]
    """
    score_calculation = []
    for scr_calc in data.score_calculation:
        scr_calc[0] = field_names_to_dataclass_fields[scr_calc[0]]
        try:
            scr_calc[1] = user_input_in_bff.field_names_to_dataclass_fields[scr_calc[1]]
        except KeyError:
            raise KeyError(
                f'Field {scr_calc[1]} not found in dataclass. Please select possible values: {user_input_in_bff.field_names_to_dataclass_fields.keys()}'
            )
        score_calculation.append(scr_calc)
    return score_calculation



import random
import sys
import uuid
from copy import deepcopy
from typing import Dict, Optional

import requests
from docarray import DocumentArray
from jina.clients import Client

from now.admin.update_api_keys import update_api_keys
from now.app.base.app import JinaNOWApp
from now.constants import ACCESS_PATHS
from now.data_loading.data_loading import load_data
from now.deployment.flow import deploy_flow
from now.log import time_profiler
from now.now_dataclasses import UserInput
from now.utils.jcloud.helpers import get_flow_id


@time_profiler
def run(
    app_instance: JinaNOWApp,
    user_input: UserInput,
    **kwargs,
):
    """
    This function will run the backend of the app. Specifically, it will:
    - Load the data
    - Set up the flow dynamically and get the environment variables
    - Deploy the flow
    - Index the data
    :param app_instance: The app instance
    :param user_input: The user input
    :param kwargs: Additional arguments
    :return:
    """
    print_callback = kwargs.get('print_callback', print)

    dataset = load_data(user_input, print_callback)
    print_callback('Data loaded. Deploying the flow...')

    # Set up the app specific flow
    app_instance.setup(user_input=user_input)

    client, gateway_port, gateway_host_internal = deploy_flow(
        flow_yaml=app_instance.flow_yaml
    )

    # TODO at the moment the scheduler is not working. So we index the data right away
    # if (
    #     user_input.deployment_type == 'remote'
    #     and user_input.dataset_type == DatasetTypes.S3_BUCKET
    #     and 'NOW_CI_RUN' not in os.environ
    # ):
    #     # schedule the trigger which will sync the bucket with the indexer once a day
    #     trigger_scheduler(user_input, gateway_host_internal)
    # else:
    # index the data right away
    print_callback('Flow deployed. Indexing the data...')
    index_docs(user_input, dataset, client, print_callback, **kwargs)

    return (
        gateway_port,
        gateway_host_internal,
    )


def trigger_scheduler(user_input, host):
    """
    This function will trigger the scheduler which will sync the bucket with the indexer once a day
    """
    print('Triggering scheduler to index data from S3 bucket')
    # check if the api_key exists. If not then create a new one
    if user_input.secured and not user_input.api_key:
        user_input.api_key = uuid.uuid4().hex
        # Also call the bff to update the api key
        for i in range(
            100
        ):  # increase the probability that all replicas get the new key
            update_api_keys(user_input.api_key, host)

    scheduler_params = {
        'flow_id': get_flow_id(host),
        'api_key': user_input.api_key,
    }
    cookies = {'st': user_input.jwt['token']}
    try:
        response = requests.post(
            'https://storefrontapi.nowrun.jina.ai/api/v1/schedule_sync',
            json=scheduler_params,
            cookies=cookies,
        )
        response.raise_for_status()
        print(
            'Scheduler triggered successfully. Scheduler will sync data from S3 bucket once a day.'
        )
    except Exception as e:
        print(f'Error while scheduling indexing: {e}')
        print(f'Indexing will not be scheduled. Please contact Jina AI support.')


def index_docs(user_input, dataset, client, print_callback, **kwargs):
    """
    Index the data right away
    """
    print_callback(f"▶ indexing {len(dataset)} documents")
    params = {'access_paths': ACCESS_PATHS}
    if user_input.secured:
        params['jwt'] = user_input.jwt
    call_flow(
        client=client,
        dataset=dataset,
        max_request_size=user_input.app_instance.max_request_size,
        parameters=deepcopy(params),
        return_results=False,
        **kwargs,
    )
    print_callback('⭐ Success - your data is indexed')


@time_profiler
def call_flow(
    client: Client,
    dataset: DocumentArray,
    max_request_size: int,
    endpoint: str = '/index',
    parameters: Optional[Dict] = None,
    return_results: Optional[bool] = False,
    **kwargs,
):
    request_size = estimate_request_size(dataset, max_request_size)

    response = client.post(
        on=endpoint,
        request_size=request_size,
        inputs=dataset,
        show_progress=True,
        parameters=parameters,
        continue_on_error=True,
        prefetch=100,
        on_done=kwargs.get('on_done', None),
        on_error=kwargs.get('on_error', None),
        on_always=kwargs.get('on_always', None),
    )

    if return_results:
        return response


def estimate_request_size(index, max_request_size):
    if len(index) > 30:
        sample = random.sample(index, 30)
    else:
        sample = index
    size = sum([sys.getsizeof(x.content) for x in sample]) / 30
    max_size = 50_000
    request_size = max(min(max_request_size, int(max_size / size)), 1)
    return request_size



class Parameters:
    host: str = None
    port: str = None
    index_fields: str = None  # concatenated string of fields for indexing
    data: str = None
    code: str = None
    state: str = None
    secured: bool = False
    top_k: int = 9


S3_DEMO_PATH = 'https://jina-now-demo.s3.eu-central-1.amazonaws.com/'

ds_set = [
    'nft-monkey',
    'deepfashion',
    'nih-chest-xrays',
    'stanford-cars',
    'bird-species',
    'best-artworks',
    'geolocation-geoguessr',
    'rock-lyrics',
    'pop-lyrics',
    'rap-lyrics',
    'indie-lyrics',
    'metal-lyrics',
    'tumblr-gifs',
    'tumblr-gifs-10k',
]

BUTTONS = {
    'twitter': """
                <a href="https://twitter.com/share?ref_src=twsrc%5Etfw" class="twitter-share-button"
                Tweet
                </a>
                <script async src="https://platform.twitter.com/widgets.js" charset="utf-8"></script>
            """,
    'linkedin': """
                <a href="https://www.linkedin.com/sharing/share-offsite/?url=https://now.jina.ai"
                class="linkedin-share-button"
                rel="noreferrer noopener" when using target="_blank">
                </a>
                <script src="https://platform.linkedin.com/in.js" type="text/javascript">lang: en_US</script>
                <script type="IN/Share" data-url="https://now.jina.ai"></script>
            """,
    'facebook': """
                <a href="https://www.facebook.com/sharer.php?u=https://now.jina.ai" class="facebook-share-button"
                rel="noreferrer noopener" when using target="_blank">
                </a>
                <div id="fb-root"></div>
                <script async defer crossorigin="anonymous"
                src="https://connect.facebook.net/en_GB/sdk.js#xfbml=1&version=v14.0" nonce="kquhy3fp"></script>
                <div class="fb-share-button" data-href="https://now.jina.ai" data-layout="button" data-size="small">
                <a target="_blank"
                 href="https://www.facebook.com/sharer/sharer.php?u=https%3A%2F%2Fnow.jina.ai%2F&amp;src=sdkpreparse"
                  class="fb-xfbml-parse-ignore">Share</a></div>
            """,
}

SSO_COOKIE = 'st'
AVATAR_COOKIE = 'AvatarUrl'
TOKEN_COOKIE = 'JinaNOW_Token'



from now.constants import DEFAULT_FLOW_NAME


def create_jcloud_name(flow_name: str) -> str:
    return (
        flow_name + '-' + DEFAULT_FLOW_NAME
        if flow_name and flow_name != DEFAULT_FLOW_NAME
        else DEFAULT_FLOW_NAME
    )



import json
import logging
import os

from now.now_dataclasses import UserInput

logger = logging.getLogger(__name__)

ENV_PREFIX = 'JINA_NOW_'

# server
DEFAULT_WORKERS = 1
DEFAULT_PORT = 8080
DEFAULT_BACKLOG = 2048

# debug flag
DEFAULT_DEBUG = True

# logging
DEFAULT_LOGGING_LEVEL = 'DEBUG'
DEFAULT_LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'basic': {
            'class': 'logging.Formatter',
            'format': '%(asctime)s  %(name)-30s  %(levelname)8s  ::  %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'basic',
            'stream': 'ext://sys.stdout',
        }
    },
    'root': {'level': 'DEBUG', 'handlers': ['console']},
}


user_input_in_bff = UserInput()


def init_user_input_in_bff():
    global user_input_in_bff
    try:
        with open(os.path.join(os.path.expanduser('~'), 'user_input.json'), 'r') as f:
            user_input_dict = json.load(f)
        for attr_name, prev_value in user_input_in_bff.__dict__.items():
            setattr(
                user_input_in_bff,
                attr_name,
                user_input_dict.get(attr_name, prev_value),
            )
    except FileNotFoundError:
        print('Could not find user input file in BFF')
        print(f'used path: {os.path.join(os.path.expanduser("~"), "user_input.json")}')
        print('but this can be okay')


init_user_input_in_bff()



from __future__ import annotations, print_function, unicode_literals

from docarray import DocumentArray
from docarray.typing import Image, Text, Video
from pydantic import BaseModel

from now.utils.common.helpers import BetterEnum


class DemoDatasetNames(BetterEnum):
    BEST_ARTWORKS = 'team-now/best-artworks'
    TLL = 'team-now/totally-looks-like'
    BIRD_SPECIES = 'team-now/bird-species'
    STANFORD_CARS = 'team-now/stanford-cars'
    DEEP_FASHION = 'team-now/deepfashion'
    POP_LYRICS = 'team-now/pop-lyrics'
    TUMBLR_GIFS_10K = 'team-now/tumblr-gifs-10k'
    ES_ONLINE_SHOP_50 = 'team-now/extracted-data-online-shop-50-flat'


class DemoDataset(BaseModel):
    name: str
    display_name: str
    index_fields: str  # To be removed once the app works with all index fields
    info: str
    source: str

    def get_data(self, *args, **kwargs) -> DocumentArray:
        return DocumentArray.pull(self.name)


AVAILABLE_DATASETS = {
    Image: [
        DemoDataset(
            name=DemoDatasetNames.BEST_ARTWORKS,
            display_name='🖼 artworks (≈8K docs)',
            index_fields='image',
            info='A collection of artworks of the 50 most influential artists of all time.',
            source='https://www.kaggle.com/datasets/ikarus777/best-artworks-of-all-time',
        ),
        DemoDataset(
            name=DemoDatasetNames.TLL,
            display_name='👬 totally looks like (≈12K docs)',
            index_fields='image',
            info='A collection of 6_016 image-pairs from '
            'the wild to cover the diversity at which humans operate',
            source='https://sites.google.com/view/totally-looks-like-dataset',
        ),
        DemoDataset(
            name=DemoDatasetNames.BIRD_SPECIES,
            display_name='🦆 birds (≈12K docs)',
            index_fields='image',
            info='A collection of images of birds containing 12_000 images of 200 species of birds.',
            source='https://www.vision.caltech.edu/datasets/cub_200_2011/',
        ),
        DemoDataset(
            name=DemoDatasetNames.STANFORD_CARS,
            display_name='🚗 cars (≈16K docs)',
            index_fields='image',
            info='A collection of images of cars containing 16_185 images of 196 classes of cars.',
            source='https://ai.stanford.edu/~jkrause/cars/car_dataset.html',
        ),
        DemoDataset(
            name=DemoDatasetNames.DEEP_FASHION,
            display_name='👕 fashion (≈53K docs)',
            index_fields='image',
            info='A collection of images of fashion items containing 53_000 images of 50 classes of fashion '
            'items ranging from well-posed shop images to unconstrained consumer photos.',
            source='https://mmlab.ie.cuhk.edu.hk/projects/DeepFashion.html',
        ),
    ],
    Text: [
        DemoDataset(
            name=DemoDatasetNames.POP_LYRICS,
            display_name='🎤 pop lyrics (200K docs)',
            index_fields='lyrics',
            info='A collection of pop song lyrics containing 200_000 song lyrics',
            source='NA',
        ),
    ],
    Video: [
        DemoDataset(
            name=DemoDatasetNames.TUMBLR_GIFS_10K,
            display_name='🎦 tumblr gifs subset (10K gifs)',
            index_fields='video',
            info='A collection of gifs from tumblr containing a (10_000) animated GIFs and sentences '
            'describing visual content of the animated GIFs',
            source='https://raingo.github.io/TGIF-Release/',
        ),
    ],
}



import os.path
import tempfile

from jina.clients import Client

from now.deployment.deployment import deploy_wolf
from now.log import time_profiler
from now.utils.jcloud.helpers import write_flow_file


@time_profiler
def deploy_flow(
    flow_yaml: str,
):
    """Deploy a Flow on JCloud, Kubernetes, or using Jina Orchestration"""
    # TODO create tmpdir top level and pass it down
    with tempfile.TemporaryDirectory() as tmpdir:
        # hack we don't know if the flow yaml is a path or a string
        if type(flow_yaml) == dict:
            flow_file = os.path.join(tmpdir, 'flow.yml')
            write_flow_file(flow_yaml, flow_file)
            flow_yaml = flow_file

        flow = deploy_wolf(path=flow_yaml)
        host = flow.endpoints['gateway (grpc)']
        client = Client(host=host)

        # host & port
        gateway_port = None
        gateway_host_internal = host
    return (
        client,
        gateway_port,
        gateway_host_internal,
    )



import base64
from time import sleep
from typing import Any, Dict, Optional

import filetype
from docarray import Document
from docarray.document.mixins.helper import _is_datauri, _to_datauri, _uri_to_blob
from pydantic import BaseModel, Field, root_validator


class BaseRequestModel(BaseModel):
    jwt: Optional[Dict[str, Any]] = Field(
        default={},
        description="Dictionary with key 'token' which maps to Jina Cloud token value."
        " To be passed when the flow is secure",
        example={'token': '<your jina cloud token>'},
    )
    api_key: Optional[str] = Field(
        default=None,
        description='Used to authenticate machines',
        example='<your api key>',
    )

    class Config:
        allow_mutation = False
        case_sensitive = False
        arbitrary_types_allowed = True


class ModalityModel(BaseModel):
    uri: Optional[str] = Field(
        default=None,
        description='URI of the file or data URI',
        example='https://example.com/image.jpg',
    )
    text: Optional[str] = Field(default=None, description='Text', example='cute cats')
    blob: Optional[str] = Field(
        default=None, description='Base64 encoded `utf-8` str format', example='xxx'
    )

    @root_validator(pre=True)
    def validate_only_one_exists(cls, values):
        # Get the names of all fields that are set (i.e. have a non-None value)
        set_fields = [name for name, value in values.items() if value]
        if len(set_fields) != 1:
            raise ValueError(f"Only one of {set_fields} can be set.")
        return values

    @property
    def content(self):
        for field_name in self.__fields_set__:
            field_value = getattr(self, field_name)
            if field_value:
                return field_value

    def to_html(self, title: str = '', disable_to_datauri: bool = False) -> str:
        """Converts the ModalityModel to HTML.

        :param title: Title of the figure (field name usually).
        :param disable_to_datauri: If True, the image is not converted to datauri.
        """
        if self.uri or self.blob:
            if self.uri:
                if _is_datauri(self.uri) or disable_to_datauri:
                    src = self.uri
                else:
                    # try downloading 5 times
                    src = None
                    for _ in range(5):
                        try:
                            _blob = _uri_to_blob(self.uri, timeout=10)
                            src = _to_datauri(
                                Document(uri=self.uri).mime_type,
                                _blob,
                                'utf-8',
                                False,
                                binary=True,
                            )
                            break
                        except:
                            sleep(1)
                            continue
                    if src is None:
                        src = self.uri
            elif self.blob:
                base64_decoded = base64.b64decode(self.blob.encode('utf-8'))
                file_ending = filetype.guess(base64_decoded)
                if not file_ending:
                    raise ValueError(
                        f'Could not guess file type of blob {self.blob}. '
                        f'Please provide a valid file type.'
                    )
                src = f'data:{file_ending.mime}/{file_ending.extension};base64,{self.blob}'
            html = f'<img src="{src}" alt="missing" style="max-width: 100px; max-height: 100px;">'

            if title:
                html = (
                    '<figure>' + html + f'<figcaption>({title})</figcaption></figure>'
                )
        elif self.text is not None:
            html = f'<p><b>{self.text}</b>' + (f' ({title})' if title else '') + '</p>'
        return html



from docarray import Document
from fastapi import APIRouter, HTTPException

from now.executor.gateway.bff.app.v1.models.admin import (
    UpdateApiKeysRequestModel,
    UpdateEmailsRequestModel,
)
from now.executor.gateway.bff.app.v1.models.shared import BaseRequestModel
from now.executor.gateway.bff.app.v1.routers.helper import jina_client_post

router = APIRouter()


@router.post(
    "/updateUserEmails",
    summary='update user emails during runtime',
)
async def update_user_email(data: UpdateEmailsRequestModel):
    """
    Update the list of emails for the security executor
    """
    await jina_client_post(
        request_model=data,
        docs=Document(),
        endpoint='/admin/updateUserEmails',
        parameters={'user_emails': data.user_emails},
    )


@router.post(
    "/updateApiKeys",
    summary='update api keys during runtime',
)
async def update_api_keys(data: UpdateApiKeysRequestModel):
    """
    Update the list of api keys for the security executor
    """
    await jina_client_post(
        request_model=data,
        docs=Document(),
        endpoint='/admin/updateApiKeys',
        parameters={'api_keys': data.api_keys},
    )


@router.post(
    "/getStatus",
    summary='Get status of the flow during runtime',
)
async def get_host_status(data: BaseRequestModel):
    """
    Get the status of the host in the request body
    """
    try:
        await jina_client_post(
            request_model=data,
            docs=Document(),
            endpoint='/dry_run',
        )
    except Exception as e:
        raise HTTPException(status_code=404, detail=str(e))
    return 'SUCCESS'



import json
import os.path

import cowsay
import requests
from docarray import DocumentArray
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Column, Table

from now import run_backend
from now.compare.compare_flows import compare_flows_for_queries
from now.constants import DEMO_NS, FLOW_STATUS
from now.deployment import deployment
from now.dialog import configure_user_input, maybe_prompt_user


def stop_now(**kwargs):
    _result, flow_id, cluster = get_flow_status(action='delete', **kwargs)
    if _result is not None and _result['status']['phase'] == FLOW_STATUS:
        deployment.terminate_wolf(flow_id)
        from hubble import Client

        cookies = {'st': Client().token}
        requests.delete(
            f'https://storefrontapi.nowrun.jina.ai/api/v1/schedule_sync/{flow_id}',
            cookies=cookies,
        )
    cowsay.cow(f'remote Flow `{cluster}` removed')


def start_now(**kwargs):
    user_input = configure_user_input(**kwargs)
    app_instance = user_input.app_instance
    # Only if the deployment is remote and the demo examples is available for the selected app
    # Should not be triggered for CI tests
    if app_instance.is_demo_available(user_input):
        gateway_host_grpc = f'grpcs://{DEMO_NS.format(user_input.dataset_name.split("/")[-1])}.dev.jina.ai'
    else:
        (
            gateway_port,
            gateway_host_grpc,
        ) = run_backend.run(app_instance, user_input, **kwargs)
    gateway_host_http = gateway_host_grpc.replace('grpc', 'http')
    bff_url = f'{gateway_host_http}/api/v1/search-app/docs'
    playground_url = f'{gateway_host_http}/playground'

    _generate_info_table(
        gateway_host_http, gateway_host_grpc, bff_url, playground_url, user_input
    )
    return {
        'bff': bff_url,
        'playground': playground_url,
        'host_http': gateway_host_http,
        'host_grpc': gateway_host_grpc,
        'secured': user_input.secured,
    }


def get_docarray(dataset):
    if os.path.exists(dataset):
        print(f'Loading queries from {dataset}')
        return DocumentArray.load_binary(dataset)
    print(f'Pulling queries from {dataset}')
    da = DocumentArray.pull(name=dataset, show_progress=True)
    if 'NOW_CI_RUN' in os.environ:
        da = da[:21]
    return da


def compare_flows(**kwargs):
    if not 'flow_ids' in kwargs:
        path_score_calculation = maybe_prompt_user(
            [
                {
                    'type': 'input',
                    'name': 'path_score_calculation',
                    'message': 'Path to the json file mapping flow ID to a list of score calculation configurations (optional):',
                }
            ],
            'path_score_calculation',
            **kwargs,
        )
        if path_score_calculation:
            with open(path_score_calculation) as fp:
                cluster_ids_2_score_calculation = json.load(fp)

            flow_ids = list(cluster_ids_2_score_calculation.keys())
            flow_ids_http_score_calculation = [
                (flow_id, f'https://{flow_id}-http.wolf.jina.ai', score_calculation)
                for flow_id in flow_ids
                for score_calculation in cluster_ids_2_score_calculation[flow_id]
            ]
    if 'flow_ids' in kwargs or not path_score_calculation:
        flow_ids = maybe_prompt_user(
            [
                {
                    'type': 'input',
                    'name': 'flow_ids',
                    'message': 'Enter comma-separated the flow names to compare:',
                }
            ],
            'flow_ids',
            **kwargs,
        )
        flow_ids_http_score_calculation = [
            (cluster_id, f'https://{cluster_id}-http.wolf.jina.ai', [])
            for cluster_id in flow_ids.split(',')
        ]

    dataset = maybe_prompt_user(
        [
            {
                'type': 'input',
                'name': 'dataset',
                'message': 'Path to the DocArray with the queries in multi-modal format',
            }
        ],
        'dataset',
        **kwargs,
    )
    da = get_docarray(dataset)
    if not da[0].is_multimodal:
        raise ValueError(
            f'The DocArray {dataset} is not a multimodal DocumentArray.'
            f'Please check documentation https://docarray.jina.ai/fundamentals/dataclass/construct/'
        )

    limit = maybe_prompt_user(
        [
            {
                'type': 'input',
                'name': 'limit',
                'message': 'Enter the number of results to compare:',
            }
        ],
        'limit',
        **kwargs,
    )
    limit = int(limit)

    disable_to_datauri = maybe_prompt_user(
        [
            {
                'type': 'list',
                'choices': [
                    {'name': '⛔ no', 'value': False},
                    {'name': '✅ yes', 'value': True},
                ],
                'name': 'disable_to_datauri',
                'message': 'Disable loading to DataURI (makes the files smaller but also not self-contained)?',
            }
        ],
        'disable_to_datauri',
        **kwargs,
    )

    results_per_table = maybe_prompt_user(
        [
            {
                'type': 'input',
                'name': 'results_per_table',
                'message': 'Enter the number of results shown per table (default is 20):',
            }
        ],
        'results_per_table',
        **kwargs,
    )
    results_per_table = int(results_per_table) if results_per_table else 20

    compare_flows_for_queries(
        da=da,
        flow_ids_http_score_calculation=flow_ids_http_score_calculation,
        limit=limit,
        results_per_table=results_per_table,
        disable_to_datauri=disable_to_datauri,
    )


def get_flow_status(action, **kwargs):
    choices = []
    # Add all remote Flows that exists with the namespace `nowapi`
    alive_flows = deployment.list_all_wolf()
    for flow_details in alive_flows:
        choices.append(flow_details['name'])
    if len(choices) == 0:
        cowsay.cow(f'nothing to {action}')
        return
    else:
        questions = [
            {
                'type': 'list',
                'name': 'cluster',
                'message': f'Which cluster do you want to {action}?',
                'choices': choices,
            }
        ]
        cluster = maybe_prompt_user(questions, 'cluster', **kwargs)

    flow = [x for x in alive_flows if x['name'] == cluster][0]
    flow_id = flow['id']
    _result = deployment.status_wolf(flow_id)
    if _result is None:
        print(f'❎ Flow not found in JCloud. Likely, it has been deleted already')
    return _result, flow_id, cluster


def _generate_info_table(
    gateway_host_http, gateway_host_grpc, bff_url, playground_url, user_input
):
    info_table = Table(
        'Attribute',
        Column(header="Value", overflow="fold"),
        show_header=False,
        box=box.SIMPLE,
        highlight=True,
    )
    info_table.add_row('Host (HTTPS)', gateway_host_http)
    info_table.add_row('Host (GRPCS)', gateway_host_grpc)
    info_table.add_row('API docs', bff_url)
    if user_input.secured and user_input.api_key:
        info_table.add_row('API Key', user_input.api_key)
    info_table.add_row('Playground', playground_url)
    console = Console()
    console.print(
        Panel(
            info_table,
            title=f':tada: Search app is NOW ready!',
            expand=False,
        )
    )



import os
import traceback
from collections import namedtuple
from time import sleep
from typing import Any, Dict, List, Optional, Tuple

import boto3
from docarray import Document, DocumentArray
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

from now.constants import DatasetTypes
from now.executor.abstract.auth import (
    SecurityLevel,
    get_auth_executor_class,
    secure_request,
)
from now.executor.indexer.elastic.es_converter import (
    convert_doc_map_to_es,
    convert_es_results_to_matches,
    convert_es_to_da,
)
from now.executor.indexer.elastic.es_query_building import (
    build_es_queries,
    generate_score_calculation,
    process_filter,
)

FieldEmbedding = namedtuple(
    'FieldEmbedding',
    ['encoder', 'embedding_size', 'fields'],
)

Executor = get_auth_executor_class()

TIMEOUT = 60


class NOWElasticIndexer(Executor):
    """
    NOWElasticIndexer indexes Documents into an Elasticsearch instance. To do this,
    it uses helper functions from es_converter, converting documents to and from the accepted Elasticsearch
    format. It also uses the score calculation to combine the scores of different fields/encoders,
    allowing multi-modal documents to be indexed and searched with multi-modal queries.
    """

    def __init__(
        self,
        document_mappings: List[Tuple[str, int, List[str]]],
        metric: str = 'cosine',
        limit: int = 10,
        max_values_per_tag: int = 10,
        es_mapping: Dict = None,
        es_config: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ):
        """
        :param document_mappings: list of FieldEmbedding tuples that define which encoder
            encodes which fields, and the embedding size of the encoder.
        :param metric: Distance metric type. Can be 'euclidean', 'inner_product', or 'cosine'
        :param limit: Number of results to get for each query document in search
        :param max_values_per_tag: Maximum number of values per tag
        :param es_mapping: Mapping for new index. If none is specified, this will be
            generated from `document_mappings` and `metric`.
        :param hosts: host configuration of the Elasticsearch node or cluster
        :param es_config: Elasticsearch cluster configuration object
        :param index_name: ElasticSearch Index name used for the storage
        """

        super().__init__(*args, **kwargs)
        self.metric = metric
        self.limit = limit
        self.max_values_per_tag = max_values_per_tag
        self._check_env_vars()
        self.hosts = os.getenv('ES_HOSTS', 'http://localhost:9200')
        self.api_key = os.getenv('ES_API_KEY', 'TestApiKey')
        self.index_name = os.getenv('ES_INDEX_NAME', 'now-index')
        self.query_to_curated_ids = {}
        self.filters_val_dict = {}
        self.document_mappings = [FieldEmbedding(*dm) for dm in document_mappings]
        self.encoder_to_fields = {
            document_mapping.encoder: document_mapping.fields
            for document_mapping in self.document_mappings
        }
        self.es_config = es_config or {'verify_certs': False}
        self.es_mapping = es_mapping or self.generate_es_mapping()
        self.es = Elasticsearch(
            hosts=self.hosts,
            api_key=self.api_key,
            **self.es_config,
            ssl_show_warn=False,
        )
        self._do_health_check()
        if not self.es.indices.exists(index=self.index_name):
            self.es.indices.create(index=self.index_name, mappings=self.es_mapping)
        else:
            self.es.indices.put_mapping(index=self.index_name, body=self.es_mapping)

    def _check_env_vars(self):
        while not all(
            var in os.environ for var in ['ES_HOSTS', 'ES_INDEX_NAME', 'ES_API_KEY']
        ):
            timeout_counter = 0
            if timeout_counter < TIMEOUT:
                timeout_counter += 5
                self.logger.info('Environment variables not set yet. Waiting...')
                sleep(5)
            else:
                self.logger.error(
                    'Elasticsearch environment variables not set after 60 seconds. Exiting...'
                )
                raise Exception('Elasticsearch environment variables not set')

    def _do_health_check(self):
        """Checks that Elasticsearch is up and running with state 'yellow'. The default timeout
        on the health check is 30 seconds."""
        while True:
            try:
                self.es.cluster.health(wait_for_status='yellow')
                break
            except Exception:
                self.logger.info(traceback.format_exc())

    def generate_es_mapping(self) -> Dict:
        """Creates Elasticsearch mapping for the defined document fields."""
        es_mapping = {
            'properties': {
                'id': {'type': 'keyword'},
            }
        }

        for field in self.user_input.index_fields:
            if self.user_input.index_field_candidates_to_modalities[field] == 'text':
                es_mapping['properties'][
                    f"{self.user_input.field_names_to_dataclass_fields[field]}"
                ] = {'type': 'text', 'analyzer': 'standard'}

        if self.user_input.filter_fields:
            es_mapping['properties']['tags'] = {'type': 'object', 'properties': {}}
            for field in self.user_input.filter_fields:
                es_mapping['properties']['tags']['properties'][field] = {
                    'type': 'keyword'
                }

        for encoder, embedding_size, fields in self.document_mappings:
            for field in fields:
                es_mapping['properties'][f'{field}-{encoder}'] = {
                    'properties': {
                        f'embedding': {
                            'type': 'dense_vector',
                            'dims': str(embedding_size),
                            'similarity': self.metric,
                            'index': 'true',
                        }
                    }
                }
        return es_mapping

    def _handle_no_docs_map(self, docs: DocumentArray):
        if docs and len(self.encoder_to_fields) == 1:
            return {list(self.encoder_to_fields.keys())[0]: docs}
        else:
            return {}

    @secure_request(on='/index', level=SecurityLevel.USER)
    def index(
        self,
        docs_map: Dict[str, DocumentArray] = None,  # encoder to docarray
        docs: Optional[DocumentArray] = None,
        **kwargs,
    ) -> DocumentArray:
        """
        Index new `Document`s by adding them to the Elasticsearch index.

        :param docs_map: map of encoder to DocumentArray
        :param docs: DocumentArray to index
        :return: empty `DocumentArray`.
        """
        if docs_map is None:
            docs_map = self._handle_no_docs_map(docs)
            if len(docs_map) == 0:
                return DocumentArray()
        aggregate_embeddings(docs_map)
        es_docs = convert_doc_map_to_es(
            docs_map, self.index_name, self.encoder_to_fields
        )
        success, _ = bulk(self.es, es_docs)
        self.es.indices.refresh(index=self.index_name)
        if success:
            self.logger.info(
                f'Inserted {success} documents into Elasticsearch index {self.index_name}'
            )
        self.update_tags()
        return DocumentArray([])

    @secure_request(on='/search', level=SecurityLevel.USER)
    def search(
        self,
        docs_map: Dict[str, DocumentArray] = None,
        parameters: dict = {},
        docs: Optional[DocumentArray] = None,
        **kwargs,
    ):
        """Perform traditional bm25 + vector search.

        Search can be performed with candidate filtering. Filters are a triplet (column,operator,value).
        More than a filter can be applied during search. Therefore, conditions for a filter are specified as a list triplets.
        Each triplet contains:
            - field: Field used to filter.
            - operator: Binary operation between two values. Some supported operators include `['>','<','=','<=','>=']`.
            - value: value used to compare a candidate.

        :param docs_map: map of encoder to DocumentArray
        :param parameters: dictionary of options for searching.
            Keys accepted:
                - 'filter' (dict): The filtering conditions on document tags
                - 'limit' (int): Number of matches to get per Document, default 100.
                - 'get_score_breakdown' (bool): Wether to return the score breakdown, i.e. the scores of each
                    field+encoder combination/comparison.
                - 'score_calculation' (List[List]): list of tuples of (query_field, document_field, matching_method,
                    linear_weight) to show how to calculate the score. Note, that the matching_method is the name of the
                    encoder or `bm25`.
        :param docs: DocumentArray to search
        """
        if docs_map is None:
            docs_map = self._handle_no_docs_map(docs)
            if len(docs_map) == 0:
                return DocumentArray()
        aggregate_embeddings(docs_map)

        filter = parameters.get('filter', {})
        limit = parameters.get('limit', self.limit)
        get_score_breakdown = parameters.get('get_score_breakdown', False)
        score_calculation = parameters.get('score_calculation', None)
        if not score_calculation:
            score_calculation = generate_score_calculation(
                docs_map, self.encoder_to_fields
            )

        es_queries = build_es_queries(
            docs_map=docs_map,
            get_score_breakdown=get_score_breakdown,
            score_calculation=score_calculation,
            metric=self.metric,
            filter=filter,
            query_to_curated_ids=self.query_to_curated_ids,
        )
        for doc, query in es_queries:
            result = self.es.search(
                index=self.index_name,
                query=query,
                source=True,
                size=limit,
            )['hits']['hits']
            doc.matches = convert_es_results_to_matches(
                query_doc=doc,
                es_results=result,
                get_score_breakdown=get_score_breakdown,
                metric=self.metric,
                score_calculation=score_calculation,
            )
            doc.tags.pop('embeddings')
            for c in doc.chunks:
                c.embedding = None
        results = DocumentArray(list(zip(*es_queries))[0])

        if (
            parameters.get('create_temp_link', False)
            and self.user_input.dataset_type == DatasetTypes.S3_BUCKET
        ):
            self._create_temporary_links(results)

        return results

    def _create_temporary_links(self, docs: DocumentArray):
        """For every match, it replaces the URI with a temporary link such that no credentials are needed for access."""

        def _create_temp_link(d: Document) -> Document:
            if (
                not d.text
                and not d.blob
                and isinstance(d.uri, str)
                and d.uri.startswith('s3://')
            ):
                session = boto3.session.Session(
                    aws_access_key_id=self.user_input.aws_access_key_id,
                    aws_secret_access_key=self.user_input.aws_secret_access_key,
                    region_name=self.user_input.aws_region_name,
                )
                s3_client = session.client('s3')
                bucket_name = d.uri.split('/')[2]
                path_s3 = '/'.join(d.uri.split('/')[3:])
                temp_url = s3_client.generate_presigned_url(
                    'get_object',
                    Params={'Bucket': bucket_name, 'Key': path_s3},
                    ExpiresIn=3600,
                )
                d.uri = temp_url
            return d

        for d in docs['@mc,mcc']:
            _create_temp_link(d)

    @secure_request(on='/update', level=SecurityLevel.USER)
    def update(self, docs: DocumentArray, **kwargs) -> DocumentArray:
        """
        TODO: implement update endpoint, eg. update ES docs with new embeddings etc.
        """
        raise NotImplementedError()

    @secure_request(on='/list', level=SecurityLevel.USER)
    def list(self, parameters: dict = {}, **kwargs):
        """List all indexed documents.

        Note: this implementation is naive and does not
        consider the default maximum documents in a page returned by `Elasticsearch`.
        Should be addressed in future with `scroll`.

        :param parameters: dictionary with limit and offset
        - offset (int): number of documents to skip
        - limit (int): number of retrieved documents
        """
        limit = int(parameters.get('limit', self.limit))
        offset = int(parameters.get('offset', 0))
        try:
            result = self.es.search(
                index=self.index_name, size=limit, from_=offset, query={'match_all': {}}
            )['hits']['hits']
        except Exception:
            result = None
            self.logger.info(traceback.format_exc())
        if result:
            return convert_es_to_da(result, get_score_breakdown=False)
        else:
            return DocumentArray()

    @secure_request(on='/count', level=SecurityLevel.USER)
    def count(self, parameters: dict = {}, **kwargs):
        """Count all indexed documents.

        Note: this implementation is naive and does not
        consider the default maximum documents in a page returned by `Elasticsearch`.
        Should be addressed in future with `scroll`.

        :param parameters: dictionary with limit and offset
        - offset (int): number of documents to skip
        - limit (int): number of retrieved documents
        """
        limit = int(parameters.get('limit', self.limit))
        offset = int(parameters.get('offset', 0))
        try:
            result = self.es.search(
                index=self.index_name, size=limit, from_=offset, query={'match_all': {}}
            )['hits']['hits']
        except Exception:
            result = []
            self.logger.info(traceback.format_exc())
        return DocumentArray([Document(tags={'count': len(result)})])

    @secure_request(on='/delete', level=SecurityLevel.USER)
    def delete(self, parameters: dict = {}, **kwargs):
        """
        Endpoint to delete documents from an index. Either delete documents by filter condition
        or by specifying a list of document IDs.

        :param parameters: dictionary with filter conditions or list of IDs to select
            documents for deletion.
        """
        search_filter = parameters.get('filter', None)
        ids = parameters.get('ids', None)
        if search_filter:
            es_search_filter = {
                'query': {'bool': {'filter': process_filter(search_filter)}}
            }
            try:
                resp = self.es.delete_by_query(
                    index=self.index_name, body=es_search_filter
                )
                self.es.indices.refresh(index=self.index_name)
                self.update_tags()
            except Exception:
                self.logger.info(traceback.format_exc())
                raise
        elif ids:
            resp = {'deleted': 0}
            try:
                for id in ids:
                    r = self.es.delete(index=self.index_name, id=id)
                    self.es.indices.refresh(index=self.index_name)
                    resp['deleted'] += r['result'] == 'deleted'
            except Exception as e:
                self.logger.info(traceback.format_exc(), e)
        else:
            raise ValueError('No filter or IDs provided for deletion.')
        if resp:
            self.logger.info(
                f"Deleted {resp['deleted']} documents in Elasticsearch index {self.index_name}"
            )
        return DocumentArray()

    @secure_request(on='/filters', level=SecurityLevel.USER)
    def filters(self, **kwargs):
        """
        Endpoint to get all filter fields in the indexer and their possible values.
        """
        return DocumentArray([Document(tags={'filters': self.filters_val_dict})])

    @secure_request(on='/curate', level=SecurityLevel.USER)
    def curate(self, parameters: dict = {}, **kwargs):
        """
        This endpoint is only relevant for text queries.
        It defines the top results as a list of IDs for
        each query, and stores these as dictionary items.
        `query_to_filter` sent in the `parameters` should
        have the following format:
        {
            'query_to_filter': {
                'query1': [
                    {'uri': {'$eq': 'uri1'}},
                    {'tags__internal_id': {'$eq': 'id1'}},
                ],
                'query2': [
                    {'uri': {'$eq': 'uri2'}},
                    {'tags__color': {'$eq': 'red'}},
                ],
            }
        }
        """
        search_filter = parameters.get('query_to_filter', None)
        if search_filter:
            self.update_curated_ids(search_filter)
        else:
            raise ValueError('No filter provided for curating.')

    def update_curated_ids(self, search_filter):
        for query, filters in search_filter.items():
            if query not in self.query_to_curated_ids:
                self.query_to_curated_ids[query] = []
            for filter in filters:
                es_query = {'query': {'bool': {'filter': process_filter(filter)}}}

                resp = self.es.search(index=self.index_name, body=es_query, size=100)
                self.es.indices.refresh(index=self.index_name)
                ids = [r['_id'] for r in resp['hits']['hits']]
                self.query_to_curated_ids[query] += [
                    id for id in ids if id not in self.query_to_curated_ids[query]
                ]

    def update_tags(self):
        """
        The indexer keeps track of which tags are indexed and what their possible
        values are, which is stored in self.filters_val_dict. This method queries the
        elasticsearch index for the current es_mapping to find the current tags on all
        indexed documents. It then queries elasticsearch for an aggregation of all values
        inside this field, and updates the self.filters_val_dict dictionary with tags as keys,
        and values as values in the dictionary.
        """
        es_mapping = self.es.indices.get_mapping(index=self.index_name)
        tag_categories = (
            es_mapping.get(self.index_name, {})
            .get('mappings', {})
            .get('properties', {})
            .get('tags', {})
            .get('properties', {})
        )
        tag_categories = {
            tag: map
            for tag, map in tag_categories.items()
            if tag in self.user_input.filter_fields
        }
        aggs = {'aggs': {}, 'size': 0}
        for tag, map in tag_categories.items():
            for tag_type, extension in [
                ['text', '.keyword'],
                ['keyword', ''],
                ['float', ''],
            ]:
                if map['type'] == tag_type:
                    aggs['aggs'][tag] = {
                        'terms': {'field': f'tags.{tag}{extension}', 'size': 100}
                    }

        try:
            if not aggs['aggs']:
                return
            result = self.es.search(index=self.index_name, body=aggs)
            aggregations = result['aggregations']
            updated_tags = {}
            for tag, agg in aggregations.items():
                updated_tags[tag] = [bucket['key'] for bucket in agg['buckets']]
            self.filters_val_dict = updated_tags
        except Exception:  # noqa
            self.logger.info(traceback.format_exc())


def aggregate_embeddings(docs_map: Dict[str, DocumentArray]):
    """Aggregate embeddings of cc level to c level.

    :param docs_map: a dictionary of `DocumentArray`s, where the key is the embedding space aka encoder name.
    """
    for docs in docs_map.values():
        for doc in docs:
            for c in doc.chunks:
                if c.chunks.embeddings is not None:
                    c.embedding = c.chunks.embeddings.mean(axis=0)
                if c.chunks[0].text or not c.uri:
                    c.content = c.chunks[0].content
                c.chunks = DocumentArray()



import base64
from typing import Dict, List

import requests
import streamlit as st
from docarray import Document, DocumentArray
from docarray.score import NamedScore

from now.constants import NOWGATEWAY_BFF_PORT

from .constants import Parameters


def get_query_params() -> Parameters:
    query_parameters = st.experimental_get_query_params()
    parameters = Parameters()
    for key, val in query_parameters.items():
        if val:
            setattr(
                parameters,
                key,
                val[0],
            )
    return parameters


def call_flow(url_host, data, endpoint):
    st.session_state.search_count += 1
    response = requests.post(
        url_host, json=data, headers={"Content-Type": "application/json; charset=utf-8"}
    )

    try:
        if endpoint == 'suggestion':
            docs = DocumentArray.from_json(response.content)
        elif endpoint == 'search':
            docs = DocumentArray()
            # todo: use multimodal doc in the future

            for response_json in response.json():
                chunks = []
                for name, type_to_content in sorted(response_json['fields'].items()):
                    chunk = Document(**type_to_content)
                    if chunk.blob:
                        base64_bytes = chunk.blob.encode('utf-8')
                        chunk.blob = base64.decodebytes(base64_bytes)

                    chunks.append(chunk)
                doc = Document(
                    id=response_json['id'],
                    tags=response_json['tags'],
                    chunks=chunks,
                )
                for metric, value in response_json['scores'].items():
                    doc.scores[metric] = NamedScore(value=value['value'])
                docs.append(doc)
    except Exception:  # noqa
        try:
            json_response = response.json()

            if response.status_code == 401:
                st.session_state.error_msg = json_response['detail']
            else:
                st.session_state.error_msg = json_response['message'].replace('"', "'")
        except Exception:  # noqa
            st.session_state.error_msg = response.text
        return None

    st.session_state.error_msg = None

    if endpoint in ['search', 'suggestion']:
        return docs
    else:
        return response.json()


def multimodal_search(
    query_field_values_modalities: List[Dict],
    jwt,
    top_k=None,
    filter_dict=None,
    endpoint='search',
):
    params = get_query_params()
    url_host = f"http://localhost:{NOWGATEWAY_BFF_PORT}/api/v1/search-app/{endpoint}"

    updated_dict = {}
    if filter_dict:
        updated_dict = {k: v for k, v in filter_dict.items() if v != 'All'}
    data = {
        'limit': top_k if top_k else params.top_k,
        'filters': updated_dict,
        'score_calculation': list(st.session_state.score_calculation.values()),
        'get_score_breakdown': st.session_state.show_score_breakdown,
    }
    if endpoint == 'suggestion':
        data['text'] = query_field_values_modalities[0]['value']
    elif endpoint == 'search':
        data['query'] = query_field_values_modalities
        data['create_temp_link'] = True
        # in case the jwt is none, no jwt will be sent. This is the case when no authentication is used for that flow
    if jwt:
        data['jwt'] = jwt
    return call_flow(url_host, data, endpoint)



from collections import defaultdict

from docarray import Document
from fastapi import APIRouter, Body

from now.executor.gateway.bff.app.settings import user_input_in_bff
from now.executor.gateway.bff.app.v1.models.info import (
    CountRequestModel,
    CountResponseModel,
    EncoderToDataclassFieldsModsResponseModel,
    FiltersResponseModel,
)
from now.executor.gateway.bff.app.v1.models.search import SuggestionRequestModel
from now.executor.gateway.bff.app.v1.models.shared import BaseRequestModel
from now.executor.gateway.bff.app.v1.routers.helper import jina_client_post

router = APIRouter()

suggestion_examples = {
    'working_text': {
        'summary': 'A working example: get suggestions for a text query',
        'description': 'A working example which can be tried out. Get autocomplete suggestions for a text query.',
        'value': {
            'text': 'cute ca',
        },
    },
    'dummy': {
        'summary': 'A dummy example',
        'description': 'A dummy example,  do not run. For parameter reference only.',
        'value': {
            'jwt': {'token': '<your token>'},
            'api_key': '<your api key>',
            'text': 'cute cats',
        },
    },
}


@router.post(
    '/suggestion',
    summary='Get auto complete suggestion for query',
)
async def suggestion(data: SuggestionRequestModel = Body(examples=suggestion_examples)):
    suggest_doc = Document(text=data.text)
    docs = await jina_client_post(
        endpoint='/suggestion',
        docs=suggest_doc,
        request_model=data,
        target_executor=r'\Aautocomplete_executor\Z',
    )
    return docs.to_dict()


@router.post(
    '/filters', summary='Get all filters in the indexer and their possible values'
)
async def get_tags(data: BaseRequestModel) -> FiltersResponseModel:
    response = await jina_client_post(
        request_model=data,
        docs=Document(),
        endpoint='/filters',
        target_executor=r'\Aindexer\Z',
    )
    return FiltersResponseModel(filters=response[0].tags['filters'])


@router.post(
    '/count', summary='Get the count of the total number of documents in the indexer'
)
async def get_count(data: CountRequestModel) -> CountResponseModel:
    response = await jina_client_post(
        request_model=data,
        docs=Document(),
        endpoint='/count',
        target_executor=r'\Aindexer\Z',
        parameters={'limit': data.limit},
    )
    return CountResponseModel(number_of_docs=response[0].tags['count'])


@router.post('/encoder_to_dataclass_fields_mods', include_in_schema=False)
async def get_index_fields_dict() -> EncoderToDataclassFieldsModsResponseModel:
    index_fields_dict = defaultdict(dict)
    for index_field_raw, encoders in user_input_in_bff.model_choices.items():
        index_field = index_field_raw.replace('_model', '')
        dataclass_field = user_input_in_bff.field_names_to_dataclass_fields[index_field]
        modality = user_input_in_bff.index_field_candidates_to_modalities[index_field]
        for encoder in encoders:
            index_fields_dict[encoder][dataclass_field] = modality
    return EncoderToDataclassFieldsModsResponseModel(
        encoder_to_dataclass_fields_mods=index_fields_dict
    )



from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

from docarray import Document, DocumentArray

from now.utils.docarray.helpers import get_chunk_by_field_name

metrics_mapping = {
    'cosine': 'cosineSimilarity',
    'l2_norm': 'l2norm',
}


def generate_score_calculation(
    docs_map: Dict[str, DocumentArray],
    encoder_to_fields: Dict[str, Union[List[str], str]],
) -> List[List]:
    """
    Generate score calculation from document mappings.

    :param docs_map: dictionary mapping encoder to DocumentArray.
    :param encoder_to_fields: dictionary mapping encoder to fields.
    :return: a list of score calculation, each of which is a tuple of
        (query_field, document_field, encoder, linear_weight).
        score calculation would then be for example:
        [('query_text', 'title', 'clip', 1.0)]
    """
    score_calculation = []
    for executor_name, da in docs_map.items():
        first_doc = da[0]
        field_names = first_doc._metadata['multi_modal_schema'].keys()
        try:
            document_fields = encoder_to_fields[executor_name]
        except KeyError as e:
            raise KeyError(
                f'Documents are not encoded with same encoder as query. executor_name: {executor_name}, encoder_to_fields: {encoder_to_fields}'
            ) from e
        for field_name in field_names:
            chunk = get_chunk_by_field_name(first_doc, field_name)
            if chunk.chunks.embeddings is None and chunk.embedding is None:
                continue
            for document_field in document_fields:
                score_calculation.append(
                    [
                        field_name,
                        document_field,
                        executor_name,
                        1,
                    ]
                )

    return score_calculation


def build_es_queries(
    docs_map,
    get_score_breakdown: bool,
    score_calculation: List[Tuple],
    metric: Optional[str] = 'cosine',
    filter: dict = {},
    query_to_curated_ids: Dict[str, list] = {},
) -> Dict:
    """
    Build script-score query used in Elasticsearch. To do this, we extract
    embeddings from the query document and pass them in the script-score
    query together with the fields to search on in the Elasticsearch index.
    The query document will be returned with all of its embeddings as tags with
    their corresponding field+encoder as key.

    :param docs_map: dictionary mapping encoder to DocumentArray.
    :param get_score_breakdown: whether to return the score breakdown for matches.
        For this function, this parameter determines whether to return the embeddings
        of a query document.
    :param score_calculation: list of nested lists containing (query_field, document_field, matching_method, linear_weight) which define
        how to calculate the score. Note, that the matching_method is the name of the encoder or `bm25`.
    :param metric: metric to use for vector search.
    :param filter: dictionary of filters to apply to the search.
    :param query_to_curated_ids: dictionary mapping query text to list of curated ids.
    :return: a dictionary containing query and filter.
    """
    queries = {}
    pinned_queries = {}
    docs = {}
    sources = {}
    script_params = defaultdict(dict)
    for executor_name, da in docs_map.items():
        for doc in da:
            if doc.id not in docs:
                docs[doc.id] = doc
                docs[doc.id].tags['embeddings'] = {}

            if doc.id not in queries:
                queries[doc.id] = get_default_query(
                    doc,
                    score_calculation,
                    filter,
                )
                pinned_queries[doc.id] = get_pinned_query(
                    doc,
                    query_to_curated_ids,
                )

                if any(
                    _matching_method == 'bm25'
                    for (_, _, _matching_method, _) in score_calculation
                ):
                    sources[doc.id] = '1.0 + _score / (_score + 10.0)'
                else:
                    sources[doc.id] = '1.0'

            for (
                query_field,
                document_field,
                matching_method,
                linear_weight,
            ) in get_scores(executor_name, score_calculation):
                field_doc = get_chunk_by_field_name(doc, query_field)
                if get_score_breakdown:
                    docs[doc.id].tags['embeddings'][
                        f'{query_field}-{matching_method}'
                    ] = field_doc.embedding

                query_string = f'params.query_{query_field}_{executor_name}'
                document_string = f'{document_field}-{matching_method}'

                sources[
                    doc.id
                ] += f" + {float(linear_weight)}*{metrics_mapping[metric]}({query_string}, '{document_string}.embedding')"

                script_params[doc.id][
                    f'query_{query_field}_{executor_name}'
                ] = field_doc.embedding

    es_queries = []

    for doc_id, query in queries.items():
        script_score = {
            'script_score': {
                'query': {
                    'bool': query['bool'],
                },
                'script': {
                    'source': sources[doc_id],
                    'params': script_params[doc_id],
                },
            },
        }
        if pinned_queries[doc_id]:
            query_json = {'pinned': pinned_queries[doc_id]['pinned']}
            query_json['pinned']['organic'] = script_score
        else:
            query_json = script_score
        es_queries.append((docs[doc_id], query_json))
    return es_queries


def get_default_query(
    doc: Document,
    score_calculation: List[Tuple],
    filter: Dict = {},
):
    query = {
        'bool': {
            'should': [
                {'match_all': {}},
            ],
        },
    }

    # build bm25 part
    for (query_field, index_field, matching_method, linear_weight) in score_calculation:
        if matching_method == 'bm25':
            text = get_chunk_by_field_name(doc, query_field).text
            query['bool']['should'].append(
                {
                    'multi_match': {
                        'query': text,
                        'fields': [f"{index_field}^{linear_weight}"],
                    }
                }
            )

    # add filter
    if filter:
        es_search_filter = process_filter(filter)
        query['bool']['filter'] = es_search_filter

    return query


def get_pinned_query(doc: Document, query_to_curated_ids: Dict[str, list] = {}) -> Dict:
    pinned_query = {}
    if getattr(doc, 'query_text', None):
        query_text = doc.query_text.text
        if query_text in query_to_curated_ids.keys():
            pinned_query = {'pinned': {'ids': query_to_curated_ids[query_text]}}
    return pinned_query


def process_filter(
    filter: Dict[str, Union[List[str], Dict[str, float]]]
) -> List[Dict[str, Any]]:
    es_search_filters = []
    for field, filters in filter.items():
        field = field.replace('__', '.', 1)
        es_search_filter = {}
        if isinstance(filters, list):  # must be categorical (list of terms)
            es_search_filter['terms'] = {field: filters}
        elif isinstance(filters, dict):  # must be numerical (range with operators)
            es_search_filter['range'] = {field: filters}
        else:
            raise ValueError(
                f'Filter {field}: {filters} is not a list of terms or a dictionary of ranges'
            )
        es_search_filters.append(es_search_filter)
    return es_search_filters


def get_scores(encoder, score_calculation):
    for (
        query_field,
        document_field,
        _encoder,
        linear_weight,
    ) in score_calculation:
        if encoder == _encoder:
            yield query_field, document_field, _encoder, linear_weight



import json
import os
import shutil
from time import sleep
from typing import Dict, List, Tuple

import streamlit.web.bootstrap
from jina import Gateway
from jina.enums import GatewayProtocolType
from jina.serve.runtimes.gateway import CompositeGateway
from jina.serve.runtimes.gateway.http.fastapi import FastAPIBaseGateway
from jina.serve.runtimes.gateway.http.models import JinaHealthModel
from streamlit.file_util import get_streamlit_file_path
from streamlit.web.server import Server as StreamlitServer

from now.constants import NOWGATEWAY_BFF_PORT
from now.deployment.deployment import cmd
from now.executor.gateway.hubble_report import start_base_fee_thread
from now.now_dataclasses import UserInput

cur_dir = os.path.dirname(__file__)
TIMEOUT = 60


class PlaygroundGateway(Gateway):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.streamlit_script = 'playground/playground.py'
        # copy playground/config.toml to streamlit config.toml
        streamlit_config_toml_src = os.path.join(cur_dir, 'playground', 'config.toml')
        streamlit_config_toml_dest = get_streamlit_file_path("config.toml")
        # create streamlit_config_toml_dest if it doesn't exist
        os.makedirs(os.path.dirname(streamlit_config_toml_dest), exist_ok=True)
        shutil.copyfile(streamlit_config_toml_src, streamlit_config_toml_dest)

    async def setup_server(self):
        streamlit.web.bootstrap._fix_sys_path(self.streamlit_script)
        streamlit.web.bootstrap._fix_matplotlib_crash()
        streamlit.web.bootstrap._fix_tornado_crash()
        streamlit.web.bootstrap._fix_sys_argv(self.streamlit_script, ())
        streamlit.web.bootstrap._fix_pydeck_mapbox_api_warning()
        streamlit_cmd = f'streamlit run {self.streamlit_script}'

        self.streamlit_server = StreamlitServer(
            os.path.join(cur_dir, self.streamlit_script), streamlit_cmd
        )

    async def run_server(self):
        await self.streamlit_server.start()
        streamlit.web.bootstrap._on_server_start(self.streamlit_server)
        streamlit.web.bootstrap._set_up_signal_handler(self.streamlit_server)

    async def shutdown(self):
        self.streamlit_server.stop()


class BFFGateway(FastAPIBaseGateway):
    @property
    def app(self):
        from now.executor.gateway.bff.app.app import application

        # fix to use starlette instead of FastAPI app (throws warning that "/" is used for health checks
        application.add_route(
            path='/', route=lambda: JinaHealthModel(), methods=['GET']
        )

        return application


class NOWGateway(CompositeGateway):
    """The NOWGateway assumes that the gateway has been started with http on port 8081 and grpc on port 8085.
    This is the port on which the nginx process listens. After nginx has been started,
    it will start the playground on port 8501 and the BFF on port 8080. The actual
    HTTP gateway will start on port 8082.
    Nginx is configured to route the requests in the following way:
    - /playground -> playground on port 8501
    - /api -> BFF on port 8080
    - / -> HTTP gateway on port 8082
    No rerouting is done for the grpc gateway.
    """

    def __init__(self, user_input_dict: Dict = {}, **kwargs):
        # need to update port ot 8082, as nginx will listen on 8081
        http_idx = kwargs['runtime_args']['protocol'].index(GatewayProtocolType.HTTP)
        http_port = kwargs['runtime_args']['port'][http_idx]
        grpc_idx = kwargs['runtime_args']['protocol'].index(GatewayProtocolType.GRPC)
        grpc_port = kwargs['runtime_args']['port'][grpc_idx]
        if kwargs['runtime_args']['port'][http_idx] != 8081:
            raise ValueError(
                f'Please, let http port ({http_port}) be 8081 for nginx to work'
            )
        if grpc_port in [8080, 8081, 8082, 8501]:
            raise ValueError(
                f'Please, let grpc port ({grpc_port}) be different from 8080 (BFF), '
                f'8081 (nginx), 8082 (http) and 8501 (playground)'
            )
        kwargs['runtime_args']['port'][http_idx] = 8082
        super().__init__(**kwargs)
        self.storage_dir = None
        self.authorized_jwt = None

        # Hacky method since `workspace` class variable is not available in Gateway
        try:
            self.storage_dir = [
                folder
                for folder in os.listdir('/data')
                if folder.startswith('jnamespace-')
            ]
            if len(self.storage_dir) == 0:
                self.logger.info('No storage directory found')
            else:
                self.logger.info(f'Found storage directory: {self.storage_dir}')
                self.storage_dir = self.storage_dir[0]
        except Exception as e:
            self.logger.info(f'Error while getting storage directory: {e}')

        self._check_env_vars()

        self.user_input = UserInput()
        for attr_name, prev_value in self.user_input.__dict__.items():
            setattr(
                self.user_input,
                attr_name,
                user_input_dict.get(attr_name, prev_value),
            )
        # we need to write the user input to a file so that the playground can read it; this is a workaround
        # for the fact that we cannot pass arguments to streamlit (StreamlitServer doesn't respect it)
        # we also need to do this for the BFF
        # save user_input to file in home directory of user
        with open(os.path.join(os.path.expanduser('~'), 'user_input.json'), 'w') as f:
            json.dump(self.user_input.__dict__, f)

        # remove potential clashing arguments from kwargs
        kwargs.pop("port", None)
        kwargs.pop("protocol", None)

        # note order is important
        self._add_gateway(
            BFFGateway,
            NOWGATEWAY_BFF_PORT,
            **kwargs,
        )
        self._add_gateway(
            PlaygroundGateway,
            8501,
            **kwargs,
        )

        self.setup_nginx()
        self.nginx_was_shutdown = False

        if self.storage_dir:
            if os.path.isfile(f'{self.storage_dir}/cred.json'):
                self.logger.info('Found cred.json file. Loading from it')
                with open(f'{self.storage_dir}/cred.json', 'r') as f:
                    cred_data = json.load(f)
                    self.authorized_jwt = cred_data.get('authorized_jwt', None)
            else:
                self.logger.info('No cred.json file found to load from')

        try:
            start_base_fee_thread(
                self.user_input.jwt['token'], self.authorized_jwt, self.storage_dir
            )
        except Exception as e:
            self.logger.error(f'Could not start base fee thread: {e}')

    def _check_env_vars(self):
        while 'M2M' not in os.environ:
            timeout_counter = 0
            if timeout_counter < TIMEOUT:
                timeout_counter += 5
                self.logger.info('Environment variables not set yet. Waiting...')
                sleep(5)
            else:
                self.logger.error(
                    'Gateway environment variables not set after 60 seconds. Exiting...'
                )
                raise Exception('Gateway environment variables not set')

    async def shutdown(self):
        await super().shutdown()
        if not self.nginx_was_shutdown:
            self.shutdown_nginx()
            self.nginx_was_shutdown = True

    def setup_nginx(self):
        command = [
            'nginx',
            '-c',
            os.path.join(cur_dir, '', 'nginx.conf'),
        ]
        output, error = self._run_nginx_command(command)
        self.logger.info('Nginx started')
        self.logger.info(f'nginx output: {output}')
        self.logger.info(f'nginx error: {error}')

    def shutdown_nginx(self):
        command = ['nginx', '-s', 'stop']
        output, error = self._run_nginx_command(command)
        self.logger.info('Nginx stopped')
        self.logger.info(f'nginx output: {output}')
        self.logger.info(f'nginx error: {error}')

    def _run_nginx_command(self, command: List[str]) -> Tuple[bytes, bytes]:
        self.logger.info(f'Running command: {command}')
        output, error = cmd(command)
        if error != b'':
            # on CI we need to use sudo; using NOW_CI_RUN isn't good if running test locally
            self.logger.info(f'nginx error: {error}')
            command.insert(0, 'sudo')
            self.logger.info(f'So running command: {command}')
            output, error = cmd(command)
        sleep(10)
        return output, error

    def _add_gateway(self, gateway_cls, port, protocol='http', **kwargs):
        # ignore metrics_registry since it is not copyable
        runtime_args = self._deepcopy_with_ignore_attrs(
            self.runtime_args,
            [
                'metrics_registry',
                'tracer_provider',
                'grpc_tracing_server_interceptors',
                'aio_tracing_client_interceptors',
                'tracing_client_interceptor',
            ],
        )
        runtime_args.port = [port]
        runtime_args.protocol = [protocol]
        gateway_kwargs = {k: v for k, v in kwargs.items() if k != 'runtime_args'}
        gateway_kwargs['runtime_args'] = dict(vars(runtime_args))
        gateway = gateway_cls(**gateway_kwargs)
        gateway.streamer = self.streamer
        self.gateways.insert(0, gateway)



import requests
from docarray import dataclass
from docarray.typing import Text

from now.executor.gateway.bff.app.v1.models.search import SearchRequestModel
from now.executor.gateway.bff.app.v1.routers import helper


class Client:
    """
    This is the jina NOW client which can be used to run requests against apps which are deployed using jcloud.
    """

    def __init__(self, jcloud_id, app, api_key):
        self.jcloud_id = jcloud_id
        self.app = app
        self.api_key = api_key

    def send_request_bff(self, endpoint: str, **kwargs):
        request_body = {
            "host": f'grpcs://nowapi-{self.jcloud_id}.wolf.jina.ai',
            "api_key": self.api_key,
            **kwargs,
        }
        response = requests.post(
            f'https://nowrun.jina.ai/api/v1/{endpoint}',
            json=request_body,
        )
        return response

    async def send_request(self, endpoint: str, **kwargs):
        """
        Client to run requests against a deployed flow
        """
        if endpoint != 'search':
            raise NotImplementedError('Only search endpoint is supported for now')

        @dataclass
        class DataClass:
            text_0: Text

        if 'text' in kwargs:
            query_doc = helper.field_dict_to_mm_doc(
                {'text': kwargs.pop('text')},
                data_class=DataClass,
                modalities_dict={'text': Text},
                field_names_to_dataclass_fields={'text': 'text_0'},
            )
        else:
            raise Exception('query doc is empty')
        app_request = SearchRequestModel(
            host=f'grpcs://nowapi-{self.jcloud_id}.wolf.jina.ai',
            api_key=self.api_key,
            **kwargs,
        )
        response = await helper.jina_client_post(
            app_request,
            endpoint,
            docs=query_doc,
            parameters={
                'limit': app_request.limit,
                'filter': app_request.filters,
                'score_calculation': app_request.score_calculation,
            },
        )
        return response



import signal
import sys
from collections.abc import MutableMapping


def flatten_dict(d, parent_key='', sep='__'):
    """
    This function converts a nested dictionary into a dictionary of attirbutes using '__' as a separator.
    Example:
        {'a': {'b': {'c': 1, 'd': 2}}} -> {'a__b__c': 1, 'a__b__c': 2}
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            # TODO for now, we just have string values, str(v) should be removed once we support numeric values
            items.append((new_key, str(v)))
    return dict(items)


def hide_string_chars(s):
    return ''.join(['*' for _ in range(len(s) - 4)]) + s[len(s) - 4 :] if s else None


def to_camel_case(text):
    """Convert text to camel case in great coding style"""
    s = text.replace("_", " ")
    s = s.split()
    return ''.join(i.capitalize() for i in s)


def my_handler(signum, frame, spinner):
    with spinner.hidden():
        sys.stdout.write("Program terminated!\n")
    spinner.stop()
    exit(0)


class BetterEnum:
    def __iter__(self):
        return [getattr(self, x) for x in dir(self) if ('__' not in x)].__iter__()


sigmap = {signal.SIGINT: my_handler, signal.SIGTERM: my_handler}



import logging
from dataclasses import dataclass

import boto3
import hubble

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@dataclass
class AWSProfile:
    aws_access_key_id: str
    aws_secret_access_key: str
    region: str


def get_aws_profile():
    session = boto3.Session()
    credentials = session.get_credentials()
    aws_profile = (
        AWSProfile(credentials.access_key, credentials.secret_key, session.region_name)
        if credentials
        else AWSProfile(None, None, session.region_name)
    )
    return aws_profile


@hubble.login_required
def jina_auth_login():
    pass


def get_info_hubble(user_input):
    client = hubble.Client(max_retries=None, jsonify=True)
    response = client.get_user_info()
    user_input.admin_emails = (
        [response['data']['email']] if 'email' in response['data'] else []
    )
    if not user_input.admin_emails:
        print(
            'Your hubble account is not verified. Please verify your account to deploy your flow as admin.'
        )
    user_input.jwt = {'token': client.token}
    user_input.admin_name = response['data']['name']
    return response['data'], client.token



import tempfile

from jina import Document, DocumentArray

from now.app.base.app import JinaNOWApp
from now.constants import DatasetTypes
from now.executor.abstract.auth import (
    SecurityLevel,
    get_auth_executor_class,
    secure_request,
)
from now.executor.preprocessor.s3_download import maybe_download_from_s3

Executor = get_auth_executor_class()


def move_uri(d: Document) -> Document:
    cloud_uri = d.tags.get('uri')
    if isinstance(cloud_uri, str) and cloud_uri.startswith('s3://'):
        d.uri = cloud_uri
        d.chunks[:, 'uri'] = cloud_uri
    return d


class NOWPreprocessor(Executor):
    """Applies preprocessing to documents for encoding, indexing and searching as defined by app.
    If necessary, downloads files for that from cloud bucket.
    """

    def __init__(self, max_workers: int = 15, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app: JinaNOWApp = JinaNOWApp()
        self.max_workers = max_workers

    @secure_request(on=None, level=SecurityLevel.USER)
    def preprocess(self, docs: DocumentArray, *args, **kwargs) -> DocumentArray:
        """If necessary downloads data from cloud bucket. Applies preprocessing to document as defined by apps.

        :param docs: loaded data but not preprocessed
        :return: preprocessed documents which are ready to be encoded and indexed
        """
        if len(docs) > 0 and not docs[0].chunks:
            raise ValueError(
                'Documents are not in multi modal format. Please check documentation'
                'https://docarray.jina.ai/datatypes/multimodal/'
            )
        with tempfile.TemporaryDirectory() as tmpdir:
            index_fields = []
            if self.user_input:
                for index_field in self.user_input.index_fields:
                    index_fields.append(
                        self.user_input.field_names_to_dataclass_fields[index_field]
                        if index_field
                        in self.user_input.field_names_to_dataclass_fields
                        else index_field
                    )

            if (
                self.user_input
                and self.user_input.dataset_type == DatasetTypes.S3_BUCKET
            ):
                maybe_download_from_s3(
                    docs=docs,
                    tmpdir=tmpdir,
                    user_input=self.user_input,
                    max_workers=self.max_workers,
                )

            docs = self.app.preprocess(docs)

            # As _maybe_download_from_s3 moves S3 URI to tags['uri'], need to move it back for post-processor & accurate
            # results.
            if (
                self.user_input
                and self.user_input.dataset_type == DatasetTypes.S3_BUCKET
            ):

                for d in docs:
                    for c in d.chunks:
                        # TODO please fix this hack - uri should not be in tags
                        move_uri(c)
        return docs
